from inewave.newave.modelos.curva import (
    BlocoConfiguracoesPenalizacaoCurva,
    BlocoPenalidadesViolacaoREECurva,
    BlocoCurvaSegurancaREE,
    BlocoMaximoIteracoesProcessoIterativoEtapa2,
    BlocoIteracaoAPartirProcessoIterativoEtapa2,
    BlocoToleranciaProcessoIterativoEtapa2,
    BlocoImpressaoRelatorioProcessoIterativoEtapa2,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Curva(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à curva para
    penalização por volume mínimo dos reservatórios.
    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoConfiguracoesPenalizacaoCurva,
        BlocoPenalidadesViolacaoREECurva,
        BlocoCurvaSegurancaREE,
        BlocoMaximoIteracoesProcessoIterativoEtapa2,
        BlocoIteracaoAPartirProcessoIterativoEtapa2,
        BlocoToleranciaProcessoIterativoEtapa2,
        BlocoImpressaoRelatorioProcessoIterativoEtapa2,
    ]

    @property
    def configuracoes_penalizacao(self) -> Optional[list]:
        """
        Linha de configuração das opções de penalização do
        arquivo curva.dat.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: Optional[list]
        """
        b = self.data.get_sections_of_type(BlocoConfiguracoesPenalizacaoCurva)
        if isinstance(b, BlocoConfiguracoesPenalizacaoCurva):
            return b.data
        return None

    @configuracoes_penalizacao.setter
    def configuracoes_penalizacao(self, valor: list):
        b = self.data.get_sections_of_type(BlocoConfiguracoesPenalizacaoCurva)
        if isinstance(b, BlocoConfiguracoesPenalizacaoCurva):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def custos_penalidades(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os custos para penalização em cada REE.

        - codigo_ree (`int`)
        - penalidade (`float`)

        :return: Os custos por REE em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoPenalidadesViolacaoREECurva)
        if isinstance(b, BlocoPenalidadesViolacaoREECurva):
            return b.data
        return None

    @custos_penalidades.setter
    def custos_penalidades(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoPenalidadesViolacaoREECurva)
        if isinstance(b, BlocoPenalidadesViolacaoREECurva):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def curva_seguranca(self) -> Optional[pd.DataFrame]:
        """
        Tabela da curva de segurança por REE.

        - codigo_ree (`int`)
        - data (`datetime`)
        - valor (`float`)

        :return: Os valores dos campos da linha como uma lista.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoCurvaSegurancaREE)
        if isinstance(b, BlocoCurvaSegurancaREE):
            return b.data
        return None

    @curva_seguranca.setter
    def curva_seguranca(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoCurvaSegurancaREE)
        if isinstance(b, BlocoCurvaSegurancaREE):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def maximo_iteracoes_etapa2(self) -> Optional[int]:
        """
        Número máximo de iterações no processo iterativo de 2ª
        etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.data.get_sections_of_type(
            BlocoMaximoIteracoesProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoMaximoIteracoesProcessoIterativoEtapa2):
            return b.data
        return None

    @maximo_iteracoes_etapa2.setter
    def maximo_iteracoes_etapa2(self, valor: int):
        b = self.data.get_sections_of_type(
            BlocoMaximoIteracoesProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoMaximoIteracoesProcessoIterativoEtapa2):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def iteracao_a_partir_etapa2(self) -> Optional[int]:
        """
        Iteração a partir da qual ocorre o processo iterativo de 2ª
        etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.data.get_sections_of_type(
            BlocoIteracaoAPartirProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoIteracaoAPartirProcessoIterativoEtapa2):
            return b.data
        return None

    @iteracao_a_partir_etapa2.setter
    def iteracao_a_partir_etapa2(self, valor: int):
        b = self.data.get_sections_of_type(
            BlocoIteracaoAPartirProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoIteracaoAPartirProcessoIterativoEtapa2):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def tolerancia_processo_etapa2(self) -> Optional[float]:
        """
        Tolerância para a execução do processo iterativo de etapa 2.

        :return: O valor como um float
        :rtype: Optional[float]
        """
        b = self.data.get_sections_of_type(
            BlocoToleranciaProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoToleranciaProcessoIterativoEtapa2):
            return b.data
        return None

    @tolerancia_processo_etapa2.setter
    def tolerancia_processo_etapa2(self, valor: int):
        b = self.data.get_sections_of_type(
            BlocoToleranciaProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoToleranciaProcessoIterativoEtapa2):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def impressao_relatorio_etapa2(self) -> Optional[int]:
        """
        Opção de impressão ou não do relatório do
        processo iterativo de 2ª etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.data.get_sections_of_type(
            BlocoImpressaoRelatorioProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoImpressaoRelatorioProcessoIterativoEtapa2):
            return b.data
        return None

    @impressao_relatorio_etapa2.setter
    def impressao_relatorio_etapa2(self, valor: int):
        b = self.data.get_sections_of_type(
            BlocoImpressaoRelatorioProcessoIterativoEtapa2
        )
        if isinstance(b, BlocoImpressaoRelatorioProcessoIterativoEtapa2):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.ghmaxm import GHAnos


class Ghmaxm(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à geração hidraulica máxima
    por patamar, por Submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghmax00x.out`, onde x varia conforme o
    Submercado em questão.

    """

    BLOCKS = [
        Submercado,
        GHAnos,
    ]



from inewave.nwlistcf.modelos.nwlistcfdat import (
    PeriodoImpressaoCortesEstados,
    OpcoesImpressao,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional, List


class Nwlistcfdat(SectionFile):
    """
    Armazena os dados de entrada para a execução do programa auxiliar
    NWLISTCF, existentes no arquivo `nwlistcf.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [PeriodoImpressaoCortesEstados, OpcoesImpressao]

    @property
    def mes_inicio(self) -> Optional[int]:
        """
        O mês (calendário) de início para impressão dos cortes

        :return: O mês calendário de início
        :rtype: Optional[int] | None
        """
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            return b.data[0]
        return None

    @mes_inicio.setter
    def mes_inicio(self, v: int):
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            b.data[0] = v

    @property
    def mes_fim(self) -> Optional[int]:
        """
        O mês de fim para impressão dos cortes

        :return: O mês calendário de fim
        :rtype: Optional[int] | None
        """
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            return b.data[1]
        return None

    @mes_fim.setter
    def mes_fim(self, v: int):
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            b.data[1] = v

    @property
    def imprime_cortes_ativos(self) -> Optional[int]:
        """
        O flag para impressão somente dos cortes ativos.

        :return: O valor do flag
        :rtype: Optional[int] | None
        """
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            return b.data[2]
        return None

    @imprime_cortes_ativos.setter
    def imprime_cortes_ativos(self, v: int):
        b = self.data.get_sections_of_type(PeriodoImpressaoCortesEstados)
        if isinstance(b, PeriodoImpressaoCortesEstados):
            b.data[2] = v

    @property
    def opcoes_impressao(self) -> Optional[List[int]]:
        """
        As opções de impressão selecionadas.

        :return: As opções de impressão
        :rtype: Optional[List[int]] | None
        """
        b = self.data.get_sections_of_type(OpcoesImpressao)
        if isinstance(b, OpcoesImpressao):
            return b.data
        return None

    @opcoes_impressao.setter
    def opcoes_impressao(self, v: List[int]):
        b = self.data.get_sections_of_type(OpcoesImpressao)
        if isinstance(b, OpcoesImpressao):
            b.data = v



# Rotinas de testes associadas ao arquivo restricao-eletrica.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.restricaoeletrica import (
    RegistroRE,
    RegistroREHorizPer,
    RegistroRELimFormPer,
)

from inewave.newave.restricaoeletrica import RestricaoEletrica

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.restricaoeletrica import (
    MockRE,
    MockREHorizPer,
    MockRELimFormPer,
    MockRestricaoEletrica,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_re_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockRE))
    r = RegistroRE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "1.0ger_usit(13) + ger_usih(66)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "1.0ger_usit(13) + ger_usih(66)"
    r.formula = "Teste"


def test_registro_re_horiz_per_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockREHorizPer))
    r = RegistroREHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2021, 1, 1), datetime(2021, 1, 1)]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicial == datetime(2021, 1, 1)
    r.data_inicial = datetime(2021, 2, 1)
    assert r.data_final == datetime(2021, 1, 1)
    r.data_final = datetime(2022, 11, 1)


def test_registro_rhq_lim_form_per_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockRELimFormPer))
    r = RegistroRELimFormPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2021, 1, 1),
        datetime(2021, 3, 1),
        1,
        -1.1e30,
        5000.0,
    ]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicial == datetime(2021, 1, 1)
    r.data_inicial = datetime(2021, 2, 1)
    assert r.data_final == datetime(2021, 3, 1)
    r.data_final = datetime(2021, 4, 1)
    assert r.limite_inferior == -1.1e30
    r.limite_inferior = 0
    assert r.limite_superior == 5000
    r.limite_superior = 0


def test_atributos_encontrados_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEletrica))
    with patch("builtins.open", m):
        e = RestricaoEletrica.read(ARQ_TESTE)
        assert len(e.re()) > 0
        assert len(e.re_horiz_per()) > 0
        assert len(e.re_lim_form_per()) > 0


def test_eq_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEletrica))
    with patch("builtins.open", m):
        cf1 = RestricaoEletrica.read(ARQ_TESTE)
        cf2 = RestricaoEletrica.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_restricaoeletrica():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEletrica))
    with patch("builtins.open", m):
        cf1 = RestricaoEletrica.read(ARQ_TESTE)
        cf2 = RestricaoEletrica.read(ARQ_TESTE)
        cf2.data.remove(cf1.re()[0])
        assert cf1 != cf2


def test_leitura_escrita_restricaoeletrica():
    m_leitura: MagicMock = mock_open(read_data="".join(MockRestricaoEletrica))
    with patch("builtins.open", m_leitura):
        cf1 = RestricaoEletrica.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = RestricaoEletrica.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)
from inewave.nwlistop.modelos.vevminm import VevminAnos


class Vevminm(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes às violações da meta
    de energia da vazão mínima, por Submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vevminm00x.out`, onde x varia conforme o
    Submercado em questão.

    """

    BLOCKS = [
        Submercado,
        VevminAnos,
    ]



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from typing import List, IO, Optional
from datetime import datetime
import pandas as pd  # type: ignore


class BlocoUTEExpt(Section):
    """
    Bloco com as expensões das UTEs.
    """

    FIM_BLOCO = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_uhe = Line(
            [
                IntegerField(4, 0),
                LiteralField(6, 5),
                FloatField(8, 11, 2),
                DatetimeField(size=7, starting_position=20, format="%m %Y"),
                DatetimeField(size=7, starting_position=28, format="%m %Y"),
                LiteralField(40, 37),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUTEExpt):
            return False
        bloco: BlocoUTEExpt = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame()
            df["codigo_usina"] = codigos
            df["tipo"] = tipos
            df["modificacao"] = modificacoes
            df["data_inicio"] = datas_inicio
            df["data_fim"] = datas_fim
            df["nome_usina"] = nomes
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        codigos: List[Optional[int]] = []
        nomes: List[Optional[str]] = []
        tipos: List[Optional[str]] = []
        modificacoes: List[Optional[float]] = []
        datas_inicio: List[Optional[datetime]] = []
        datas_fim: List[Optional[datetime]] = []

        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                if len(codigos) > 0:
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha_uhe.read(linha)
            codigos.append(dados[0])
            tipos.append(dados[1])
            modificacoes.append(dados[2])
            datas_inicio.append(dados[3])
            datas_fim.append(dados[4])
            nomes.append(dados[5])

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do expt.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            dados_linha = linha_lida.tolist()
            file.write(self.__linha_uhe.write(dados_linha))
        file.write(f"{BlocoUTEExpt.FIM_BLOCO}\n")



"""
========================================
Análise de modificações do modif.dat
========================================
"""


# %%
# Para realizar a análise do modif.dat será utilizado o módulo plotly
import plotly.express as px
import plotly.io as pio
import pandas as pd
from datetime import datetime

pio.templates.default = "ggplot2"

# %%
# O modif.dat é principal arquivo de entrada do NEWAVE para alterações cadastrais
# relacioanadas a usinas hidrelétricas. Também nele são encontradas algumas restrições
# físicas e operativas das usinas, que frequentemente precisam ser alteradas.
# Assim, este exemplo contém uma análise e edição nas restrições operativas
# de uma UHE. O modif.dat é o único arquivo modelado com a abordagem de registros
# que é exclusivo do modelo NEWAVE e, por isso, sua lógica é diferente dos demais.
from inewave.newave import Modif

arq_modif = Modif.read("./newave/modif.dat")


# %%
# É possível visualizar todos os registros de um mesmo tipo existentes no arquivo
# no formato de DataFrame, porém este formato é apenas para visualização.
arq_modif.usina(df=True)

# %%
# O outro formato para visualização dos registros é a extração dos objetos em si,
# que é a maneira de se realizar também a edição de seus conteúdos. Desta forma, também
# é possível aplicar filtros no registro desejado.
arq_modif.vazmint(data_inicio=datetime(2021, 2, 1))

# %%
# Para editar o arquivo modif, necessitamos frequentemente de deletar registros existentes
# e de criar novos registros. Para isto, instanciamos os modelos dos registros existentes
# no módulo *inewave* e adicionamos ao objeto do arquivo através dos métodos da classe.
from inewave.newave.modif import TURBMAXT

codigo_usina_modificada = 7

novo_registro = TURBMAXT()
novo_registro.data_inicio = datetime(2021, 2, 1)
novo_registro.turbinamento = 1000.0
arq_modif.data.add_after(
    arq_modif.usina(codigo=codigo_usina_modificada),
    novo_registro,
)

# %%
# Podemos conferir se a inserção foi feita com sucesso ao listar os registros da usina
# em questão:
registros_usinas = arq_modif.modificacoes_usina(codigo_usina_modificada)
print([r for r in registros_usinas if isinstance(r, TURBMAXT)])


# %%
# Além de edições, podemos visualizar informações contidas no modif. Por exemplo, a variação
# da cota de montante de JIRAU ao longo do tempo:
from inewave.newave.modif import CMONT

registros_cmont = [
    r for r in arq_modif.modificacoes_usina(codigo=285) if isinstance(r, CMONT)
]
datas = [r.data_inicio for r in registros_cmont]
cotas = [r.nivel for r in registros_cmont]
df_cmont = pd.DataFrame(data={"data": datas, "cota": cotas})
fig = px.line(df_cmont, x="data", y="cota")
fig



from inewave.newave.modelos.pmo import (
    BlocoConvergenciaPMO,
    BlocoEafPastTendenciaHidrolPMO,
    BlocoEafPastCfugaMedioPMO,
    BlocoConfiguracoesExpansaoPMO,
    BlocoMARSPMO,
    BlocoRiscoDeficitENSPMO,
    BlocoCustoOperacaoPMO,
    BlocoCustoOperacaoTotalPMO,
    BlocoProdutibilidadesConfiguracaoPMO,
    BlocoEnergiaArmazenadaInicialPMO,
    BlocoVolumeArmazenadoInicialPMO,
)

from inewave.newave import Pmo
from datetime import datetime

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.pmo import MockBlocoEafPastTendenciaHidrolPMO
from tests.mocks.arquivos.pmo import MockBlocoEafPastCfugaMedioPMO
from tests.mocks.arquivos.pmo import MockBlocoConvergenciaPMO
from tests.mocks.arquivos.pmo import (
    MockBlocoConfiguracoesExpansaoEntradaReservatorioPMO,
)
from tests.mocks.arquivos.pmo import MockBlocoProdutibilidadesPMO
from tests.mocks.arquivos.pmo import MockBlocoMARSPMOInicial
from tests.mocks.arquivos.pmo import MockBlocoMARSPMOFinal
from tests.mocks.arquivos.pmo import MockBlocoRiscoDeficitENSPMO
from tests.mocks.arquivos.pmo import MockBlocoCustoOperacaoPMO
from tests.mocks.arquivos.pmo import MockBlocoCustoOperacaoTotalPMO
from tests.mocks.arquivos.pmo import MockBlocoEnergiaArmazenadaInicialPMO
from tests.mocks.arquivos.pmo import MockBlocoVolumeArmazenadoInicialPMO
from tests.mocks.arquivos.pmo import MockPMO

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_eafpast_tendencia_hidrologica():
    m: MagicMock = mock_open(
        read_data="".join(MockBlocoEafPastTendenciaHidrolPMO)
    )
    b = BlocoEafPastTendenciaHidrolPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 144
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 1] == 1
    assert b.data.iloc[0, 2] == 7196.96
    assert b.data.iloc[-1, -1] == 409.13


def test_eafpast_cfuga_medio():
    m: MagicMock = mock_open(read_data="".join(MockBlocoEafPastCfugaMedioPMO))
    b = BlocoEafPastCfugaMedioPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 144
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 1] == 1
    assert b.data.iloc[0, 2] == 7196.96
    assert b.data.iloc[-1, -1] == 409.13


def test_energia_armazenada_inicial():
    m: MagicMock = mock_open(
        read_data="".join(MockBlocoEnergiaArmazenadaInicialPMO)
    )
    b = BlocoEnergiaArmazenadaInicialPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 0] == "SUDESTE"
    assert b.data.iloc[0, 1] == 16191.0
    assert b.data.iloc[0, 2] == 31.8
    assert b.data.iloc[-1, -1] == 34.3


def test_volume_armazenado_inicial():
    m: MagicMock = mock_open(
        read_data="".join(MockBlocoVolumeArmazenadoInicialPMO)
    )
    b = BlocoVolumeArmazenadoInicialPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 164
    assert b.data.shape[1] == 4
    assert b.data.iloc[1, 0] == 20
    assert b.data.iloc[1, 1] == "BATALHA"
    assert b.data.iloc[1, 2] == 414.9
    assert b.data.iloc[1, 3] == 30.7
    assert b.data.iloc[-1, -1] == 0.0


def test_convergencia():
    m: MagicMock = mock_open(read_data="".join(MockBlocoConvergenciaPMO))
    b = BlocoConvergenciaPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert len(list(b.data.index)) == 135


def test_leitura_configs():
    m: MagicMock = mock_open(
        read_data="".join(MockBlocoConfiguracoesExpansaoEntradaReservatorioPMO)
    )
    b = BlocoConfiguracoesExpansaoPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == datetime(2021, 1, 1)
    assert b.data.iloc[-1, 0] == datetime(2030, 12, 1)
    assert b.data.iloc[0, 1] == 1
    assert b.data.iloc[-1, -1] == 60


def test_leitura_retas_perdas_inicial():
    m: MagicMock = mock_open(read_data="".join(MockBlocoMARSPMOInicial))
    b = BlocoMARSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 2] == 0.074069


def test_leitura_retas_perdas_final():
    m: MagicMock = mock_open(read_data="".join(MockBlocoMARSPMOFinal))
    b = BlocoMARSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 2] == 0.074069


def test_leitura_ens_deficit():
    m: MagicMock = mock_open(read_data="".join(MockBlocoRiscoDeficitENSPMO))
    b = BlocoRiscoDeficitENSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 20
    assert b.data.iloc[0, 0] == 2021
    assert b.data.iloc[0, 1] == "SUDESTE"
    assert b.data.iloc[0, 2] == 0.0
    assert b.data.iloc[0, 3] == 0.0


def test_leitura_tabelas_custos():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoOperacaoPMO))
    b = BlocoCustoOperacaoPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 1] == 21887.91


def test_leitura_custo_total():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoOperacaoTotalPMO))
    b = BlocoCustoOperacaoTotalPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data[0] == 22323.17
    assert b.data[1] == 107.00


def test_leitura_produtibilidades():
    m: MagicMock = mock_open(read_data="".join(MockBlocoProdutibilidadesPMO))
    b = BlocoProdutibilidadesConfiguracaoPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert len(list(b.data.index)) == 1312
    assert b.data.iloc[1, 1] == 1
    assert b.data.iloc[1, 2] == 0.4483
    assert b.data.iloc[1, 3] == 0.4334
    assert b.data.iloc[1, 4] == 0.4005
    assert b.data.iloc[1, 5] == 0.4625
    assert b.data.iloc[1, 6] == 0.4886
    assert b.data.iloc[1, 7] == 2.2083
    assert b.data.iloc[1, 8] == 2.1858
    assert b.data.iloc[1, 9] == 2.2292
    assert b.data.iloc[1, 10] == 2.1382
    assert b.data.iloc[1, 11] == 2.2292
    assert b.data.iloc[1, 12] == 2.2697
    assert b.data.iloc[1, 13] == 2.1382
    assert b.data.iloc[1, 14] == 2.2292
    assert b.data.iloc[1, 15] == 2.2697


def test_atributos_encontrados_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo = Pmo.read(ARQ_TESTE)
        assert pmo.eafpast_tendencia_hidrologica is not None
        assert pmo.eafpast_cfuga_medio is not None
        assert pmo.convergencia is not None
        assert pmo.configuracoes_alteracao_potencia is not None
        assert pmo.configuracoes_entrada_reservatorio is not None
        assert pmo.configuracoes_qualquer_modificacao is not None
        assert pmo.retas_perdas_engolimento(1) is not None
        assert pmo.retas_perdas_engolimento(2) is not None
        assert pmo.risco_deficit_ens is not None
        assert pmo.custo_operacao_series_simuladas is not None
        assert pmo.custo_operacao_referenciado_primeiro_mes is not None
        assert pmo.valor_esperado_periodo_estudo is not None
        assert pmo.custo_operacao_total is not None
        assert pmo.desvio_custo_operacao_total is not None
        assert pmo.produtibilidades_equivalentes is not None
        assert pmo.energia_armazenada_inicial is not None
        assert pmo.volume_armazenado_inicial is not None


def test_atributos_nao_encontrados_pmo():
    m: MagicMock = mock_open(read_data="".join(MockBlocoConvergenciaPMO))
    with patch("builtins.open", m):
        pmo = Pmo.read(ARQ_TESTE)
        assert pmo.convergencia is not None
        assert pmo.custo_operacao_series_simuladas is None
        assert pmo.custo_operacao_total is None
        assert pmo.desvio_custo_operacao_total is None
        assert pmo.produtibilidades_equivalentes is None
        assert pmo.energia_armazenada_inicial is None
        assert pmo.volume_armazenado_inicial is None


def test_eq_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo1 = Pmo.read(ARQ_TESTE)
        pmo2 = Pmo.read(ARQ_TESTE)
        assert pmo1 == pmo2


def test_neq_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo1 = Pmo.read(ARQ_TESTE)
        pmo2 = Pmo.read(ARQ_TESTE)
        pmo2.configuracoes_alteracao_potencia
        assert pmo1 == pmo2



from inewave.nwlistcf.modelos.nwlistcfdat import (
    PeriodoImpressaoCortesEstados,
    OpcoesImpressao,
)
from inewave.nwlistcf import Nwlistcfdat

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwlistcfdat import (
    MockBlocoPeriodoImpressao,
    MockBlocoOpcoesImpressao,
    MockNwlistcfdat,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_periodo_impressao_nwlistcfdat():
    m: MagicMock = mock_open(read_data="".join(MockBlocoPeriodoImpressao))
    b = PeriodoImpressaoCortesEstados()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)
    assert b.data == [3, 60, 0]


def test_bloco_opcoes_impressao_nwlistcfdat():
    m: MagicMock = mock_open(read_data="".join(MockBlocoOpcoesImpressao))
    b = OpcoesImpressao()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)
    assert b.data == [99, None, None]


def test_atributos_encontrados_nwlistcfdat():
    m: MagicMock = mock_open(read_data="".join(MockNwlistcfdat))
    with patch("builtins.open", m):
        n = Nwlistcfdat.read(ARQ_TESTE)
        assert n.mes_inicio is not None
        assert n.mes_fim is not None
        assert n.imprime_cortes_ativos is not None
        assert n.opcoes_impressao is not None


def test_atributos_nao_encontrados_nwlistcfdat():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Nwlistcfdat.read(ARQ_TESTE)
        assert n.mes_inicio is None
        assert n.mes_fim is None
        assert n.imprime_cortes_ativos is None
        assert n.opcoes_impressao == [None, None, None]


def test_eq_nwlistcfdat():
    m: MagicMock = mock_open(read_data="".join(MockNwlistcfdat))
    with patch("builtins.open", m):
        n1 = Nwlistcfdat.read(ARQ_TESTE)
        n2 = Nwlistcfdat.read(ARQ_TESTE)
        assert n1 == n2


def test_eq_nwlistcfdat():
    m: MagicMock = mock_open(read_data="".join(MockNwlistcfdat))
    with patch("builtins.open", m):
        n1 = Nwlistcfdat.read(ARQ_TESTE)
        n1.mes_inicio = 11
        n2 = Nwlistcfdat.read(ARQ_TESTE)
        assert n1 != n2


def test_leitura_escrita_nwlistcfdat():
    m_leitura: MagicMock = mock_open(read_data="".join(MockNwlistcfdat))
    with patch("builtins.open", m_leitura):
        nc1 = Nwlistcfdat.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        nc1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        nc2 = Nwlistcfdat.read(ARQ_TESTE)
        assert nc1 == nc2



from cfinterface.components.section import Section
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import List
from typing import IO


class SecaoDadosCortesh(Section):
    """
    Registro com os dados da execução do caso existente no
    arquivo cortesh.dat
    """

    REGISTER_SIZE = 46080

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__df_ultimo_corte_por_estagio = pd.DataFrame()
        self.__df_dados_submercado = pd.DataFrame()
        self.__df_dados_uhes = pd.DataFrame()

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosCortesh):
            return False
        bloco: SecaoDadosCortesh = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    def read(self, file: IO, *args, **kwargs):
        # Leitura do primeiro registro (dados gerais)
        self.__le_primeiro_registro(file)
        # Segundo registro (ultimo registro de cortes por estagio)
        self.__le_segundo_registro(file)
        # Terceiro registro (ordens dos modelos PARP)
        self.__le_terceiro_registro(file)
        # Quarto registro (configuracoes)
        self.__le_quarto_registro(file)
        # Quinto registro (duracoes dos patamares)
        self.__le_quinto_registro(file)
        # Sexto registro (iteracao atual)
        self.__le_sexto_registro(file)
        # Setimo registro (dados da curva de aversao)
        self.__le_setimo_registro(file)
        # Oitavo registro (dados da SAR)
        self.__le_oitavo_registro(file)
        # Nono registro (dados do CVaR)
        self.__le_nono_registro(file)
        # Decimo registro (usinas hidreletricas)
        self.__le_decimo_registro(file)
        # Decimo primeiro registro (rees e submercados)
        self.__le_decimo_primeiro_registro(file)
        # Decimo segundo registro (dados diversos REEs)
        self.__le_decimo_segundo_registro(file)
        # Decimo terceiro registro (numero UHEs estagios individualizados)
        self.__le_decimo_terceiro_registro(file)
        # Decimo quarto registro (dados UHEs)
        self.__le_decimo_quarto_registro(file)

    def __le_primeiro_registro(self, file: IO):
        dados_primeiro_bloco = np.frombuffer(
            file.read(24 * 4),
            dtype=np.int32,
            count=24,
        )
        versao_nao_ofic = file.read(20).decode("utf-8").strip()
        dados_segundo_bloco = np.frombuffer(
            file.read(10 * 4),
            dtype=np.int32,
            count=10,
        )
        self.data = (
            list(dados_primeiro_bloco)
            + [versao_nao_ofic]
            + list(dados_segundo_bloco)
        )

    def __le_segundo_registro(self, file: IO):
        file.seek(1 * self.__class__.REGISTER_SIZE)
        n_estagios = (
            self.numero_estagios_pre
            + self.numero_estagios_estudo
            + self.numero_estagios_pos
        )
        self.__tamanho_segundo_registro = n_estagios
        dados_segundo_registro = np.frombuffer(
            file.read(self.__tamanho_segundo_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_segundo_registro,
        )
        self.data += list(dados_segundo_registro)

    def __le_terceiro_registro(self, file: IO):
        file.seek(2 * self.__class__.REGISTER_SIZE)
        n_estagios = (
            self.numero_estagios_pre
            + self.numero_estagios_estudo
            + self.numero_estagios_pos
        )
        self.__tamanho_terceiro_registro = n_estagios * self.numero_rees
        dados_terceiro_registro = np.frombuffer(
            file.read(self.__tamanho_terceiro_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_terceiro_registro,
        )
        self.data += list(dados_terceiro_registro)

    def __le_quarto_registro(self, file: IO):
        file.seek(3 * self.__class__.REGISTER_SIZE)
        n_estagios = (
            self.numero_estagios_pre
            + self.numero_estagios_estudo
            + self.numero_estagios_pos
        )
        self.__tamanho_quarto_registro = n_estagios
        dados_quarto_registro = np.frombuffer(
            file.read(self.__tamanho_quarto_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_quarto_registro,
        )
        self.data += list(dados_quarto_registro)

    def __le_quinto_registro(self, file: IO):
        file.seek(4 * self.__class__.REGISTER_SIZE)
        self.__tamanho_quinto_registro = (
            self.numero_estagios_estudo * self.numero_patamares
        )
        dados_quinto_registro = np.frombuffer(
            file.read(self.__tamanho_quinto_registro * 8),
            dtype=np.float64,
            count=self.__tamanho_quinto_registro,
        )
        self.data += list(dados_quinto_registro)

    def __le_sexto_registro(self, file: IO):
        file.seek(5 * self.__class__.REGISTER_SIZE)
        self.__tamanho_sexto_registro = 1
        dados_sexto_registro = np.frombuffer(
            file.read(self.__tamanho_sexto_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_sexto_registro,
        )
        self.data += list(dados_sexto_registro)

    def __le_setimo_registro(self, file: IO):
        file.seek(6 * self.__class__.REGISTER_SIZE)
        self.__tamanho_setimo_registro = (
            self.numero_rees * (self.numero_estagios_estudo + 1)
            if self.usa_curva_aversao
            else 0
        )
        dados_setimo_registro = (
            np.frombuffer(
                file.read(self.__tamanho_setimo_registro * 8),
                dtype=np.float64,
                count=self.__tamanho_setimo_registro,
            )
            if self.usa_curva_aversao
            else np.array([], dtype=np.float64)
        )
        self.data += list(dados_setimo_registro)

    def __le_oitavo_registro(self, file: IO):
        file.seek(7 * self.__class__.REGISTER_SIZE)
        self.__tamanho_oitavo_registro = (
            (2 + self.numero_estagios_estudo + 2 * self.numero_estagios_ano)
            if self.usa_sar
            else 0
        )
        dados_oitavo_registro = (
            list(
                np.frombuffer(
                    file.read(1 * 8),
                    dtype=np.float64,
                    count=1,
                )
            )
            + list(
                np.frombuffer(
                    file.read((self.__tamanho_oitavo_registro - 1) * 4),
                    dtype=np.int32,
                    count=self.__tamanho_oitavo_registro - 1,
                )
            )
            if self.usa_sar
            else []
        )
        self.data += list(dados_oitavo_registro)

    def __le_nono_registro(self, file: IO):
        file.seek(8 * self.__class__.REGISTER_SIZE)
        self.__tamanho_nono_registro = (
            (
                1
                + 2
                * (self.numero_estagios_estudo + 2 * self.numero_estagios_ano)
            )
            if self.usa_cvar
            else 0
        )
        dados_nono_registro = (
            list(
                np.frombuffer(
                    file.read(1 * 4),
                    dtype=np.int32,
                    count=1,
                )
            )
            + list(
                np.frombuffer(
                    file.read((self.__tamanho_nono_registro - 1) * 8),
                    dtype=np.float64,
                    count=self.__tamanho_nono_registro - 1,
                )
            )
            if self.usa_cvar
            else []
        )
        self.data += list(dados_nono_registro)

    def __le_decimo_registro(self, file: IO):
        file.seek(9 * self.__class__.REGISTER_SIZE)
        self.__tamanho_decimo_registro = 1 + 2 * self.numero_maximo_uhes
        dados_decimo_registro = np.frombuffer(
            file.read(self.__tamanho_decimo_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_decimo_registro,
        )
        self.data += list(dados_decimo_registro)

    def __le_decimo_primeiro_registro(self, file: IO):
        file.seek(10 * self.__class__.REGISTER_SIZE)
        tamanho_decimo_primeiro_registro_bloco_1 = (
            self.numero_submercados + self.numero_rees
        )
        dados_decimo_primeiro_registro_bloco_1 = np.frombuffer(
            file.read(tamanho_decimo_primeiro_registro_bloco_1 * 4),
            dtype=np.int32,
            count=tamanho_decimo_primeiro_registro_bloco_1,
        )
        tamanho_decimo_primeiro_registro_bloco_2 = (
            self.numero_total_submercados + self.numero_rees
        )
        dados_decimo_primeiro_registro_bloco_2: List[str] = []
        for i in range(tamanho_decimo_primeiro_registro_bloco_2):
            dados_decimo_primeiro_registro_bloco_2.append(
                file.read(10).decode("utf-8").strip()
            )
        tamanho_decimo_primeiro_registro_bloco_3 = (
            self.numero_total_submercados + self.numero_rees
        )
        dados_decimo_primeiro_registro_bloco_3 = np.frombuffer(
            file.read(tamanho_decimo_primeiro_registro_bloco_3 * 4),
            dtype=np.int32,
            count=tamanho_decimo_primeiro_registro_bloco_3,
        )
        self.data += list(dados_decimo_primeiro_registro_bloco_1)
        self.data += dados_decimo_primeiro_registro_bloco_2
        self.data += list(dados_decimo_primeiro_registro_bloco_3)
        self.__tamanho_decimo_primeiro_registro = (
            tamanho_decimo_primeiro_registro_bloco_1
            + tamanho_decimo_primeiro_registro_bloco_2
            + tamanho_decimo_primeiro_registro_bloco_3
        )

    def __le_decimo_segundo_registro(self, file: IO):
        file.seek(11 * self.__class__.REGISTER_SIZE)
        self.__tamanho_decimo_segundo_registro = 8
        dados_decimo_segundo_registro = np.frombuffer(
            file.read(self.__tamanho_decimo_segundo_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_decimo_segundo_registro,
        )
        self.data += list(dados_decimo_segundo_registro)

    def __le_decimo_terceiro_registro(self, file: IO):
        file.seek(12 * self.__class__.REGISTER_SIZE)
        self.__tamanho_decimo_terceiro_registro = self.mes_agregacao
        dados_decimo_terceiro_registro = np.frombuffer(
            file.read(self.__tamanho_decimo_terceiro_registro * 4),
            dtype=np.int32,
            count=self.__tamanho_decimo_terceiro_registro,
        )
        self.data += list(dados_decimo_terceiro_registro)

    def __le_decimo_quarto_registro(self, file: IO):
        tamanho_bloco_decimo_quarto_registro = (
            6
            + self.numero_estagios_pre
            + self.numero_estagios_estudo
            + self.numero_estagios_pos
        )
        self.__tamanho_decimo_quarto_registro = (
            self.numero_maximo_uhes * tamanho_bloco_decimo_quarto_registro
        )
        for i in range(self.numero_maximo_uhes):
            file.seek((13 + i) * self.__class__.REGISTER_SIZE)
            dados_decimo_quarto_registro = np.frombuffer(
                file.read(tamanho_bloco_decimo_quarto_registro * 4),
                dtype=np.int32,
                count=tamanho_bloco_decimo_quarto_registro,
            )
            self.data += list(dados_decimo_quarto_registro)

    def __offset_primeiro_registro(self) -> int:
        return 35

    def __offset_segundo_registro(self) -> int:
        return (
            self.__offset_primeiro_registro() + self.__tamanho_segundo_registro
        )

    def __offset_terceiro_registro(self) -> int:
        return (
            self.__offset_segundo_registro() + self.__tamanho_terceiro_registro
        )

    def __offset_quarto_registro(self) -> int:
        return (
            self.__offset_terceiro_registro() + self.__tamanho_quarto_registro
        )

    def __offset_quinto_registro(self) -> int:
        return self.__offset_quarto_registro() + self.__tamanho_quinto_registro

    def __offset_sexto_registro(self) -> int:
        return self.__offset_quinto_registro() + self.__tamanho_sexto_registro

    def __offset_setimo_registro(self) -> int:
        return self.__offset_sexto_registro() + self.__tamanho_setimo_registro

    def __offset_oitavo_registro(self) -> int:
        return self.__offset_setimo_registro() + self.__tamanho_oitavo_registro

    def __offset_nono_registro(self) -> int:
        return self.__offset_oitavo_registro() + self.__tamanho_nono_registro

    def __offset_decimo_registro(self) -> int:
        return self.__offset_nono_registro() + self.__tamanho_decimo_registro

    def __offset_decimo_primeiro_registro(self) -> int:
        return (
            self.__offset_decimo_registro()
            + self.__tamanho_decimo_primeiro_registro
        )

    def __offset_decimo_segundo_registro(self) -> int:
        return (
            self.__offset_decimo_primeiro_registro()
            + self.__tamanho_decimo_segundo_registro
        )

    def __offset_decimo_terceiro_registro(self) -> int:
        return (
            self.__offset_decimo_segundo_registro()
            + self.__tamanho_decimo_terceiro_registro
        )

    @property
    def versao_newave(self) -> int:
        return self.data[0]

    @versao_newave.setter
    def versao_newave(self, v: int):
        self.data[0] = v

    @property
    def tamanho_corte(self) -> int:
        return self.data[1]

    @tamanho_corte.setter
    def tamanho_corte(self, v: int):
        self.data[1] = v

    @property
    def tamanho_estado(self) -> int:
        return self.data[2]

    @tamanho_estado.setter
    def tamanho_estado(self, v: int):
        self.data[2] = v

    @property
    def numero_rees(self) -> int:
        return self.data[3]

    @numero_rees.setter
    def numero_rees(self, v: int):
        self.data[3] = v

    @property
    def numero_estagios_pre(self) -> int:
        return self.data[4]

    @numero_estagios_pre.setter
    def numero_estagios_pre(self, v: int):
        self.data[4] = v

    @property
    def numero_estagios_estudo(self) -> int:
        return self.data[5]

    @numero_estagios_estudo.setter
    def numero_estagios_estudo(self, v: int):
        self.data[5] = v

    @property
    def numero_estagios_pos(self) -> int:
        return self.data[6]

    @numero_estagios_pos.setter
    def numero_estagios_pos(self, v: int):
        self.data[6] = v

    @property
    def numero_estagios_ano(self) -> int:
        return self.data[7]

    @numero_estagios_ano.setter
    def numero_estagios_ano(self, v: int):
        self.data[7] = v

    @property
    def numero_configuracoes(self) -> int:
        return self.data[8]

    @numero_configuracoes.setter
    def numero_configuracoes(self, v: int):
        self.data[8] = v

    @property
    def numero_forwards(self) -> int:
        return self.data[9]

    @numero_forwards.setter
    def numero_forwards(self, v: int):
        self.data[9] = v

    @property
    def numero_patamares(self) -> int:
        return self.data[10]

    @numero_patamares.setter
    def numero_patamares(self, v: int):
        self.data[10] = v

    @property
    def ano_inicio_estudo(self) -> int:
        return self.data[11]

    @ano_inicio_estudo.setter
    def ano_inicio_estudo(self, v: int):
        self.data[11] = v

    @property
    def mes_inicio_estudo(self) -> int:
        return self.data[12]

    @mes_inicio_estudo.setter
    def mes_inicio_estudo(self, v: int):
        self.data[12] = v

    @property
    def lag_maximo_gnl(self) -> int:
        return self.data[13]

    @lag_maximo_gnl.setter
    def lag_maximo_gnl(self, v: int):
        self.data[13] = v

    @property
    def mecanismo_aversao(self) -> int:
        return self.data[14]

    @mecanismo_aversao.setter
    def mecanismo_aversao(self, v: int):
        self.data[14] = v

    @property
    def numero_submercados(self) -> int:
        return self.data[15]

    @numero_submercados.setter
    def numero_submercados(self, v: int):
        self.data[15] = v

    @property
    def numero_total_submercados(self) -> int:
        return self.data[16]

    @numero_total_submercados.setter
    def numero_total_submercados(self, v: int):
        self.data[16] = v

    @property
    def usa_curva_aversao(self) -> int:
        return self.data[17]

    @usa_curva_aversao.setter
    def usa_curva_aversao(self, v: int):
        self.data[17] = v

    @property
    def usa_sar(self) -> int:
        return self.data[18]

    @usa_sar.setter
    def usa_sar(self, v: int):
        self.data[18] = v

    @property
    def usa_cvar(self) -> int:
        return self.data[19]

    @usa_cvar.setter
    def usa_cvar(self, v: int):
        self.data[19] = v

    @property
    def considera_no_zero_calculo_zinf(self) -> int:
        return self.data[20]

    @considera_no_zero_calculo_zinf.setter
    def considera_no_zero_calculo_zinf(self, v: int):
        self.data[20] = v

    @property
    def mes_agregacao(self) -> int:
        return self.data[21]

    @mes_agregacao.setter
    def mes_agregacao(self, v: int):
        self.data[21] = v

    @property
    def numero_maximo_uhes(self) -> int:
        return self.data[22]

    @numero_maximo_uhes.setter
    def numero_maximo_uhes(self, v: int):
        self.data[22] = v

    @property
    def considera_afluencia_anual(self) -> int:
        return self.data[23]

    @considera_afluencia_anual.setter
    def considera_afluencia_anual(self, v: int):
        self.data[23] = v

    @property
    def versao_nao_oficial(self) -> str:
        return self.data[24]

    @versao_nao_oficial.setter
    def versao_nao_oficial(self, v: str):
        self.data[24] = v

    @property
    def tipo_penalizacao_curva(self) -> int:
        return self.data[25]

    @tipo_penalizacao_curva.setter
    def tipo_penalizacao_curva(self, v: int):
        self.data[25] = v

    @property
    def mes_penalizacao_curva(self) -> int:
        return self.data[26]

    @mes_penalizacao_curva.setter
    def mes_penalizacao_curva(self, v: int):
        self.data[26] = v

    @property
    def opcao_parpa(self) -> int:
        return self.data[27]

    @opcao_parpa.setter
    def opcao_parpa(self, v: int):
        self.data[27] = v

    @property
    def tipo_agregacao_caso(self) -> int:
        return self.data[28]

    @tipo_agregacao_caso.setter
    def tipo_agregacao_caso(self, v: int):
        self.data[28] = v

    @property
    def estagio_individualizado_inicial(self) -> int:
        return self.data[29]

    @estagio_individualizado_inicial.setter
    def estagio_individualizado_inicial(self, v: int):
        self.data[29] = v

    @property
    def estagio_individualizado_final(self) -> int:
        return self.data[30]

    @estagio_individualizado_final.setter
    def estagio_individualizado_final(self, v: int):
        self.data[30] = v

    @property
    def tamanho_registro_individualizado(self) -> int:
        return self.data[31]

    @tamanho_registro_individualizado.setter
    def tamanho_registro_individualizado(self, v: int):
        self.data[31] = v

    @property
    def estagio_agregado_inicial(self) -> int:
        return self.data[32]

    @estagio_agregado_inicial.setter
    def estagio_agregado_inicial(self, v: int):
        self.data[32] = v

    @property
    def estagio_agregado_final(self) -> int:
        return self.data[33]

    @estagio_agregado_final.setter
    def estagio_agregado_final(self, v: int):
        self.data[33] = v

    @property
    def tamanho_registro_agregado(self) -> int:
        return self.data[34]

    @tamanho_registro_agregado.setter
    def tamanho_registro_agregado(self, v: int):
        self.data[34] = v

    @property
    def ultimo_registro_cortes_estagio(self) -> pd.DataFrame:
        if self.__df_ultimo_corte_por_estagio.empty:
            offset = self.__offset_primeiro_registro()
            indices = self.data[
                offset : offset + self.__tamanho_segundo_registro
            ]
            self.__df_ultimo_corte_por_estagio = pd.DataFrame(
                data={
                    "tipo_estagio": ["pre"] * self.numero_estagios_pre
                    + ["estudo"] * self.numero_estagios_estudo
                    + ["pos"] * self.numero_estagios_pos,
                    "estagio": list(range(1, self.numero_estagios_pre + 1))
                    + list(range(1, self.numero_estagios_estudo + 1))
                    + list(range(1, self.numero_estagios_pos + 1)),
                    "indice_ultimo_corte": indices,
                }
            )
        return self.__df_ultimo_corte_por_estagio

    @property
    def ordens_modelos_parp(self) -> List[int]:
        offset = self.__offset_segundo_registro()
        return self.data[offset : offset + self.__tamanho_terceiro_registro]

    @property
    def configuracoes(self) -> List[int]:
        offset = self.__offset_terceiro_registro()
        return self.data[offset : offset + self.__tamanho_quarto_registro]

    @property
    def duracoes_patamares(self) -> List[float]:
        offset = self.__offset_quarto_registro()
        return self.data[offset : offset + self.__tamanho_quinto_registro]

    @property
    def iteracao_atual(self) -> int:
        offset = self.__offset_quinto_registro()
        return self.data[offset]

    @property
    def penalidade_violacao_curva(self) -> List[float]:
        if self.usa_curva_aversao == 0:
            return []
        offset = self.__offset_sexto_registro()
        return self.data[offset : offset + self.numero_rees]

    @property
    def curva_aversao(self) -> List[float]:
        if self.usa_curva_aversao == 0:
            return []
        offset = self.__offset_sexto_registro() + self.numero_rees
        return self.data[
            offset : offset + self.numero_rees * self.numero_estagios_estudo
        ]

    @property
    def penalidade_violacao_sar(self) -> float:
        if self.usa_sar == 0:
            return 0.0
        offset = self.__offset_setimo_registro()
        return self.data[offset]

    @property
    def uso_series_condicionadas_sar(self) -> int:
        if self.usa_sar == 0:
            return 0
        offset = self.__offset_setimo_registro() + 1
        return self.data[offset]

    @property
    def flag_aplicacao_sar(self) -> List[int]:
        if self.usa_sar == 0:
            return []
        offset = self.__offset_setimo_registro() + 2
        return self.data[
            offset : offset
            + self.numero_estagios_estudo
            + 2 * self.numero_estagios_ano
        ]

    @property
    def flag_cvar(self) -> int:
        if self.usa_cvar == 0:
            return 0
        offset = self.__offset_oitavo_registro()
        return self.data[offset]

    @property
    def alfa_cvar(self) -> List[float]:
        if self.usa_cvar == 0:
            return []
        offset = self.__offset_oitavo_registro() + 1
        return self.data[
            offset : offset
            + self.numero_estagios_estudo
            + 2 * self.numero_estagios_ano
        ]

    @property
    def lambda_cvar(self) -> List[float]:
        if self.usa_cvar == 0:
            return []
        offset = (
            self.__offset_oitavo_registro()
            + 1
            + self.numero_estagios_estudo
            + 2 * self.numero_estagios_ano
        )
        return self.data[
            offset : offset
            + self.numero_estagios_estudo
            + 2 * self.numero_estagios_ano
        ]

    @property
    def numero_uhes(self) -> int:
        offset = self.__offset_nono_registro()
        return self.data[offset]

    @property
    def codigos_uhes(self) -> List[int]:
        offset = self.__offset_nono_registro() + 1
        return self.data[offset : offset + self.numero_uhes]

    @property
    def codigo_interno_ree_uhes(self) -> List[int]:
        offset = self.__offset_nono_registro() + 1 + self.numero_uhes
        return self.data[offset : offset + self.numero_uhes]

    @property
    def rees_por_submercado(self) -> List[int]:
        offset = self.__offset_decimo_registro()
        return self.data[offset : offset + self.numero_submercados]

    @property
    def codigos_internos_rees_por_submercado(self) -> List[int]:
        offset = self.__offset_decimo_registro() + self.numero_submercados
        return self.data[offset : offset + self.numero_rees]

    @property
    def nomes_rees_submercados(self) -> List[str]:
        offset = (
            self.__offset_decimo_registro()
            + self.numero_submercados
            + self.numero_rees
        )
        return self.data[
            offset : offset + self.numero_rees + self.numero_submercados
        ]

    @property
    def codigos_rees_submercados(self) -> List[int]:
        offset = (
            self.__offset_decimo_registro()
            + self.numero_submercados
            + self.numero_total_submercados
            + 2 * self.numero_rees
        )
        return self.data[
            offset : offset + self.numero_rees + self.numero_submercados
        ]

    @property
    def dados_submercados(self) -> pd.DataFrame:
        if self.__df_dados_submercado.empty:
            offset_codigos = (
                self.__offset_decimo_registro()
                + self.numero_submercados
                + self.numero_total_submercados
                + 3 * self.numero_rees
            )
            offset_nomes = (
                self.__offset_decimo_registro()
                + self.numero_submercados
                + 2 * self.numero_rees
            )
            offset_numero_rees = self.__offset_decimo_registro()
            codigos = self.data[
                offset_codigos : offset_codigos + self.numero_total_submercados
            ]
            nomes = self.data[
                offset_nomes : offset_nomes + self.numero_total_submercados
            ]
            numeros = self.data[
                offset_numero_rees : offset_numero_rees
                + self.numero_submercados
            ]
            self.__df_dados_submercado = pd.DataFrame(
                data={
                    "codigo_submercado": codigos,
                    "nome_submercado": nomes,
                    "numero_rees_submercado": numeros
                    + [0] * (len(codigos) - len(numeros)),
                }
            )
        return self.__df_dados_submercado

    @property
    def numero_coeficientes_derivacao_inexata_parpa(self) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset]

    @property
    def constante_numero_rees(self) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset + 2]

    @property
    def constante_numero_rees_ordem_maxima_parp(self) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset + 3]

    @property
    def constante_numero_submercados_patamares_carga_lag_maximo_gnl(
        self,
    ) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset + 4]

    @property
    def constante_numero_maximo_uhes(self) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset + 6]

    @property
    def constante_numero_maximo_uhes_ordem_maxima_parp(self) -> int:
        offset = self.__offset_decimo_primeiro_registro()
        return self.data[offset + 7]

    @property
    def numero_uhes_estagios_individualizados(self) -> List[int]:
        offset = self.__offset_decimo_segundo_registro()
        return self.data[offset : offset + self.mes_agregacao]

    @property
    def dados_uhes(self) -> pd.DataFrame:
        if self.__df_dados_uhes.empty:
            offset = self.__offset_decimo_terceiro_registro()
            tamanho_bloco_dados_uhes = int(
                self.__tamanho_decimo_quarto_registro / self.numero_maximo_uhes
            )
            dados = np.array(
                self.data[
                    offset : offset + self.__tamanho_decimo_quarto_registro
                ]
            ).reshape((-1, tamanho_bloco_dados_uhes))
            cols_dados = [
                "codigo_usina",
                "indice_usina",
                "posto",
                "ficticia",
                "codigo_submercado",
                "mes_agregacao",
            ]
            df = pd.DataFrame(
                dados[:, :6],
                columns=cols_dados,
            )
            rees_submercado = np.array(self.rees_por_submercado).cumsum()
            codigos_internos = self.codigos_internos_rees_por_submercado
            codigos_externos = self.codigos_rees_submercados[
                : self.numero_rees
            ]
            nomes = self.nomes_rees_submercados[: self.numero_rees]
            codigos_submercados = self.codigos_rees_submercados[
                self.numero_rees :
            ]
            nomes_submercados = self.nomes_rees_submercados[self.numero_rees :]
            df["codigo_interno_ree"] = df.apply(
                lambda linha: self.codigo_interno_ree_uhes[
                    self.codigos_uhes.index(linha["codigo_usina"])
                ],
                axis=1,
            )
            df["codigo_ree"] = df.apply(
                lambda linha: codigos_externos[
                    codigos_internos.index(linha["codigo_interno_ree"])
                ],
                axis=1,
            )
            df["nome_ree"] = df.apply(
                lambda linha: nomes[
                    codigos_internos.index(linha["codigo_interno_ree"])
                ],
                axis=1,
            )
            df["codigo_submercado"] = df.apply(
                lambda linha: codigos_submercados[
                    np.where(rees_submercado >= linha["codigo_interno_ree"])[
                        0
                    ][0]
                ],
                axis=1,
            )
            df["nome_submercado"] = df.apply(
                lambda linha: nomes_submercados[
                    codigos_submercados.index(linha["codigo_submercado"])
                ],
                axis=1,
            )
            self.__df_dados_uhes = df
        return self.__df_dados_uhes



from inewave.newave.modelos.confhd import BlocoConfUHE

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Confhd(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às
    configurações das usinas hidrelétricas.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `modif.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoConfUHE]

    @property
    def usinas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as usinas.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - posto (`int`)
        - codigo_usina_jusante (`int`)
        - ree (`int`)
        - volume_inicial_percentual (`float`)
        - usina_existente (`str`)
        - usina_modificada (`int`)
        - ano_inicio_historico (`int`)
        - ano_fim_historico (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoConfUHE)
        if isinstance(b, BlocoConfUHE):
            return b.data
        return None

    @usinas.setter
    def usinas(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoConfUHE)
        if isinstance(b, BlocoConfUHE):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



MockRHE = "RRHE;       1 ; ger_ree(1) + ener_ver_ree(1)   \n"
MockRHEHorizPer = "RHE-HORIZ-PER;        2;      2021/01;      2025/12\n"
MockRHQLsLPPEarmi = "RHE-LS-LPP-EARMI;        2;    2;     -0.3;     4000.0\n"

MockRestricaoEnergia = [
    "&RHE; cod_rest;  formula\n",
    "RHE;       1 ; ger_ree(1) + ener_ver_ree(1)   \n",
    "RHE;       2 ; ger_ree(3)\n",
    "\n",
    "&RHE-HORIZ-PER; cod_rest; data_inicial; data_final\n",
    "RHE-HORIZ-PER;        1;      2021/01;      2025/12\n",
    "RHE-HORIZ-PER;        2;      2021/01;      2025/12\n",
    "\n",
    "&RHE-LS-LPP-EARMI; cod_rest; reta; coef_ang; coef_lin\n",
    "RHE-LS-LPP-EARMI;        1;    1;      0.2;    -1000.0\n",
    "RHE-LS-LPP-EARMI;        1;    2;     -0.5;    20000.0\n",
    "RHE-LS-LPP-EARMI;        2;    1;      0.5;    -1000.0\n",
    "RHE-LS-LPP-EARMI;        2;    2;     -0.3;     4000.0\n",
]



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroRHQ(Register):
    """
    Registro que contém um cadastro de restrição de
    vazão (RHQ).
    """

    IDENTIFIER = "RHQ"
    IDENTIFIER_DIGITS = 3
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroRHQHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RHQ.
    """

    IDENTIFIER = "RHQ-HORIZ-PER"
    IDENTIFIER_DIGITS = 13
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, n: datetime):
        self.data[2] = n


class RegistroRHQLsLPPVoli(Register):
    """
    Registro que contém as retas de cada restrição
    RHQ modelada como restrição linear por partes.
    """

    IDENTIFIER = "RHQ-LS-LPP-VOLI"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            FloatField(size=16, decimal_digits=4),
            FloatField(size=16, decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def indice_reta(self) -> Optional[int]:
        return self.data[1]

    @indice_reta.setter
    def indice_reta(self, c: int):
        self.data[1] = c

    @property
    def coeficiente_angular(self) -> Optional[float]:
        return self.data[2]

    @coeficiente_angular.setter
    def coeficiente_angular(self, v: float):
        self.data[2] = v

    @property
    def coeficiente_linear(self) -> Optional[float]:
        return self.data[3]

    @coeficiente_linear.setter
    def coeficiente_linear(self, v: float):
        self.data[3] = v



# Rotinas de testes associadas ao arquivo hist-ventos.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.eolicaposto import (
    RegistroPostoVentoCadastro,
    RegistroPEEPostoVento,
)

from inewave.newave.eolicaposto import EolicaPosto

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolicaposto import (
    MockRegistroPostoCadastro,
    MockRegistroPEEPosto,
    MockEolicaPosto,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_posto_cadastro_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPostoCadastro))
    r = RegistroPostoVentoCadastro()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "NEInterior"]
    assert r.codigo_posto == 1
    r.codigo_posto = 1
    assert r.nome_posto == "NEInterior"
    r.nome_posto = "NEInterior"


def test_registro_pee_posto_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEEPosto))
    r = RegistroPEEPostoVento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1]
    assert r.codigo_posto == 1
    r.codigo_posto = 2
    assert r.codigo_pee == 1
    r.codigo_pee = 2


def test_atributos_encontrados_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockEolicaPosto))
    with patch("builtins.open", m):
        e = EolicaPosto.read(ARQ_TESTE)
        assert len(e.posto_vento_cad()) > 0
        assert len(e.pee_posto()) > 0


def test_eq_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockEolicaPosto))
    with patch("builtins.open", m):
        cf1 = EolicaPosto.read(ARQ_TESTE)
        cf2 = EolicaPosto.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockEolicaPosto))
    with patch("builtins.open", m):
        cf1 = EolicaPosto.read(ARQ_TESTE)
        cf2 = EolicaPosto.read(ARQ_TESTE)
        cf2.data.remove(cf1.posto_vento_cad()[0])
        assert cf1 != cf2


def test_leitura_escrita_eolicaposto():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolicaPosto))
    with patch("builtins.open", m_leitura):
        cf1 = EolicaPosto.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = EolicaPosto.read(ARQ_TESTE)
        assert cf1 == cf2



from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.modif import (
    USINA,
    VOLMIN,
    VOLMAX,
    NUMCNJ,
    NUMMAQ,
    VAZMIN,
    CFUGA,
    CMONT,
    VMAXT,
    VMINT,
    VMINP,
    VAZMINT,
    VAZMAXT,
    TURBMAXT,
    TURBMINT,
)

from datetime import datetime
from typing import TypeVar, List, Optional, Union
import pandas as pd  # type: ignore


class Modif(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às alterações nas
    configurações das usinas hidroelétricas.
    """

    T = TypeVar("T")

    REGISTERS = [
        USINA,
        VOLMIN,
        VOLMAX,
        NUMCNJ,
        NUMMAQ,
        VAZMIN,
        CFUGA,
        CMONT,
        VMAXT,
        VMINT,
        VMINP,
        VAZMINT,
        VAZMAXT,
        TURBMAXT,
        TURBMINT,
    ]

    def usina(
        self,
        codigo: Optional[int] = None,
        nome: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[USINA, List[USINA], pd.DataFrame]]:
        """
        Obtém um registro que define a usina modificada.

        :param codigo: código da usina modificada
        :type codigo: int | None
        :param nome: nome da usina
        :type nome: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`USINA` | list[:class:`USINA`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                USINA, codigo=codigo, nome=nome
            )

    def volmin(
        self,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[VOLMIN, List[VOLMIN], pd.DataFrame]]:
        """
        Obtém um registro que define um volume mínimo.

        :param volume: volume mínimo
        :type volume: float | None
        :param unidade: unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VOLMIN` | list[:class:`VOLMIN`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VOLMIN, volume=volume, unidade=unidade
            )

    def volmax(
        self,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[VOLMAX, List[VOLMAX], pd.DataFrame]]:
        """
        Obtém um registro que define um volume máximo.

        :param volume: volume máximo
        :type volume: float | None
        :param unidade: unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VOLMAX` | list[:class:`VOLMAX`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VOLMAX, volume=volume, unidade=unidade
            )

    def numcnj(
        self,
        numero: Optional[int] = None,
        df: bool = False,
    ) -> Optional[Union[NUMCNJ, List[NUMCNJ], pd.DataFrame]]:
        """
        Obtém um registro que um número de conjuntos de máquinas.

        :param numero: o número
        :type numero: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`NUMCNJ` | list[:class:`NUMCNJ`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(NUMCNJ, numero=numero)

    def nummaq(
        self,
        conjunto: Optional[int] = None,
        numero_maquinas: Optional[int] = None,
        df: bool = False,
    ) -> Optional[Union[NUMMAQ, List[NUMMAQ], pd.DataFrame]]:
        """
        Obtém um registro que um número de máquinas por
            conjunto de máquinas.

        :param conjunto: o conjunto
        :type conjunto: int | None
        :param numero_maquinas: o número de máquinas
        :type numero_maquinas: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`NUMMAQ` | list[:class:`NUMMAQ`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                NUMMAQ, conjunto=conjunto, numero_maquinas=numero_maquinas
            )

    def vazmin(
        self,
        vazao: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[VAZMIN, List[VAZMIN], pd.DataFrame]]:
        """
        Obtém um registro que define uma vazão mínima.

        :param vazao: vazão mínima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMIN` | list[:class:`VAZMIN`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(VAZMIN, vazao=vazao)

    def cfuga(
        self,
        data_inicio: Optional[datetime] = None,
        nivel: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[CFUGA, List[CFUGA], pd.DataFrame]]:
        """
        Obtém um registro que define o nível do canal de fuga.

        :param data_inicio: data de início da validade do nível
        :type data_inicio: datetime | None
        :param nivel: o nível
        :type nivel: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`CFUGA` | list[:class:`CFUGA`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                CFUGA, data_inicio=data_inicio, nivel=nivel
            )

    def cmont(
        self,
        data_inicio: Optional[datetime] = None,
        nivel: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[CMONT, List[CMONT], pd.DataFrame]]:
        """
        Obtém um registro que define o nível do canal de montante.

        :param data_inicio: data de início da validade do nível
        :type data_inicio: datetime | None
        :param nivel: o nível
        :type nivel: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`CMONT` | list[:class:`CMONT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                CMONT, data_inicio=data_inicio, nivel=nivel
            )

    def vmaxt(
        self,
        data_inicio: Optional[datetime] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[VMAXT, List[VMAXT], pd.DataFrame]]:
        """
        Obtém um registro que define o volume máximo por período.

        :param data_inicio: data de início da validade do volume
        :type data_inicio: datetime | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMAXT` | list[:class:`VMAXT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VMAXT, data_inicio=data_inicio, volume=volume, unidade=unidade
            )

    def vmint(
        self,
        data_inicio: Optional[datetime] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[VMINT, List[VMINT], pd.DataFrame]]:
        """
        Obtém um registro que define o volume mínimo por período.

        :param data_inicio: data de início da validade do volume
        :type data_inicio: datetime | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMINT` | list[:class:`VMINT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VMINT, data_inicio=data_inicio, volume=volume, unidade=unidade
            )

    def vminp(
        self,
        data_inicio: Optional[datetime] = None,
        volume: Optional[float] = None,
        unidade: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[VMINP, List[VMINP], pd.DataFrame]]:
        """
        Obtém um registro que define o volume mínimo para penalidade.

        :param data_inicio: data de início da validade do volume
        :type data_inicio: datetime | None
        :param volume: o volume
        :type volume: float | None
        :param unidade: a unidade do volume
        :type unidade: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VMINP` | list[:class:`VMINP`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VMINP, data_inicio=data_inicio, volume=volume, unidade=unidade
            )

    def vazmint(
        self,
        data_inicio: Optional[datetime] = None,
        vazao: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[VAZMINT, List[VAZMINT], pd.DataFrame]]:
        """
        Obtém um registro que define a vazão mínima por período.

        :param data_inicio: data de início da validade do volume
        :type data_inicio: datetime | None
        :param vazao: a vazão mínima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMINT` | list[:class:`VAZMINT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VAZMINT, data_inicio=data_inicio, vazao=vazao
            )

    def vazmaxt(
        self,
        data_inicio: Optional[datetime] = None,
        vazao: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[VAZMAXT, List[VAZMAXT], pd.DataFrame]]:
        """
        Obtém um registro que define a vazão máxima por período.

        :param data_inicio: data de início da validade do volume
        :type data_inicio: datetime | None
        :param vazao: a vazão máxima
        :type vazao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`VAZMAXT` | list[:class:`VAZMAXT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                VAZMAXT, data_inicio=data_inicio, vazao=vazao
            )

    def turbmaxt(
        self,
        data_inicio: Optional[datetime] = None,
        turbinamento: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[TURBMAXT, List[TURBMAXT], pd.DataFrame]]:
        """
        Obtém um registro que define o turbinamento máximo por período.

        :param data_inicio: data de início da validade do turbinamento
        :type data_inicio: datetime | None
        :param turbinamento: o turbinamento máximo
        :type turbinamento: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`TURBMAXT` | list[:class:`TURBMAXT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                TURBMAXT, data_inicio=data_inicio, turbinamento=turbinamento
            )

    def turbmint(
        self,
        data_inicio: Optional[datetime] = None,
        turbinamento: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[TURBMINT, List[TURBMINT], pd.DataFrame]]:
        """
        Obtém um registro que define o turbinamento mínimo por período.

        :param data_inicio: data de início da validade do turbinamento
        :type data_inicio: datetime | None
        :param turbinamento: o turbinamento mínimo
        :type turbinamento: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`TURBMINT` | list[:class:`TURBMINT`] | :class:`pd.DataFrame` | None
        """
        if df:
            return self._as_df(USINA)
        else:
            return self.data.get_registers_of_type(
                TURBMINT, data_inicio=data_inicio, turbinamento=turbinamento
            )

    def modificacoes_usina(self, codigo: int) -> Optional[List[Register]]:
        """
        Filtra os registros que são associados a uma usina específica.

        :param codigo: O código da usina
        :type codigo: int
        :return: Os registros que modificam a usina
        :rtype: List[Register]
        """
        usinas = self.usina()
        if usinas is None or isinstance(usinas, USINA):
            return None
        reg_usina: List[USINA] = list(
            filter(lambda r: r.codigo == codigo, usinas)
        )
        modificacoes_usina: List[Register] = []
        if len(reg_usina) > 0:
            r = reg_usina[0].next
            while not (isinstance(r, USINA) or r.is_last):
                modificacoes_usina.append(r)
                r = r.next

        return modificacoes_usina



from typing import TypeVar, List, Optional, Union

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicasubmercado import (
    RegistroEolicaSubmercado,
    RegistroPEESubmercado,
)


class EolicaSubmercado(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao mapeamento
    de usinas eólicas e submercados.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroEolicaSubmercado, RegistroPEESubmercado]

    def eolica_submercado(
        self,
        codigo_eolica: Optional[int] = None,
        codigo_submercado: Optional[int] = None,
    ) -> Optional[
        Union[
            RegistroEolicaSubmercado,
            List[RegistroEolicaSubmercado],
        ]
    ]:
        """
        Obtém um registro que contém o mapeamento usina-submercado.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param codigo_submercado: código que especifica o submercado
        :type codigo_submercado: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaSubmercado` |
            list[:class:`RegistroEolicaSubmercado`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaSubmercado,
            codigo_eolica=codigo_eolica,
            codigo_submercado=codigo_submercado,
        )

    def pee_subm(
        self,
        codigo_pee: Optional[int] = None,
        codigo_submercado: Optional[int] = None,
    ) -> Optional[Union[RegistroPEESubmercado, List[RegistroPEESubmercado]]]:
        """
        Obtém um registro que contém o mapeamento PEE-submercado.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param codigo_submercado: código que especifica o submercado
        :type codigo_submercado: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEESubmercado` |
            list[:class:`RegistroPEESubmercado`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPEESubmercado,
            codigo_pee=codigo_pee,
            codigo_submercado=codigo_submercado,
        )



from inewave.config import MAX_LAG_ADTERM, MAX_UTES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from inewave._utils.formatacao import (
    repete_vetor,
)


class BlocoUTEAdTerm(Section):
    """
    Bloco com os despachos antecipados das UTEs por patamar.
    """

    FIM_BLOCO = " 9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ute = Line(
            [IntegerField(4, 1), LiteralField(12, 7), IntegerField(1, 21)]
        )
        self.__linha_despachos = Line([])
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUTEAdTerm):
            return False
        bloco: BlocoUTEAdTerm = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, numero_patamares: int = 3, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "codigo_usina": repete_vetor(
                        codigo_utes, self.__numero_patamares
                    ),
                    "nome_usina": repete_vetor(
                        nome_utes, self.__numero_patamares
                    ),
                    "lag": repete_vetor(lag_utes, self.__numero_patamares),
                    "patamar": np.tile(
                        np.arange(1, self.__numero_patamares + 1),
                        len(codigo_utes),
                    ),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.__numero_patamares = numero_patamares
        self.__linha_despachos = Line(
            [
                FloatField(12, 22 + 12 * i, 2)
                for i in range(self.__numero_patamares)
            ]
        )

        # Variáveis auxiliares
        codigo_utes: List[int] = []
        nome_utes: List[str] = []
        lag_utes: List[int] = []
        codigo_atual = 0
        nome_atual = ""
        lag_atual = 0
        tabela = np.zeros((MAX_UTES * MAX_LAG_ADTERM, self.__numero_patamares))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                break
            if BlocoUTEAdTerm.FIM_BLOCO in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, confere se é uma linha de UTE ou despacho
            if len(linha[:5].strip()) > 0:
                codigo_atual, nome_atual, lag_atual = self.__linha_ute.read(
                    linha
                )
                lag_atual = 0
            else:
                lag_atual += 1
                tabela[i, :] = self.__linha_despachos.read(linha)
                codigo_utes.append(codigo_atual)
                nome_utes.append(nome_atual)
                lag_utes.append(lag_atual)
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do adterm.dat não foram lidos com sucesso")
        df = self.data.copy()
        ultima_ute = 0
        for _, linha_usina in (
            self.data[["codigo_usina", "lag"]].drop_duplicates().iterrows()
        ):
            df_lag = df.loc[
                (df["codigo_usina"] == linha_usina["codigo_usina"])
                & (df["lag"] == linha_usina["lag"])
            ]
            if df_lag["codigo_usina"].iloc[0] != ultima_ute:
                ultima_ute = df_lag["codigo_usina"].iloc[0]
                max_lag_usina = df.loc[
                    df["codigo_usina"] == ultima_ute, "lag"
                ].max()
                file.write(
                    self.__linha_ute.write(
                        df_lag[["codigo_usina", "nome_usina"]]
                        .iloc[0, :]
                        .tolist()
                        + [max_lag_usina]
                    )
                )
            file.write(
                self.__linha_despachos.write(
                    df_lag.sort_values("patamar")["valor"].tolist()
                )
            )
        file.write(BlocoUTEAdTerm.FIM_BLOCO + "\n")



from inewave.newave.modelos.forwarh import SecaoDadosForwarh
from inewave.newave.forwarh import Forwarh

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch
import pytest

ARQ_TESTE = "./tests/mocks/arquivos/forwarh.dat"


def test_secao_dados_forwarh():
    r = SecaoDadosForwarh()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(fp, storage="BINARY")

    assert len(r.data) == 128


def test_atributos_nao_encontrados_forwarh():
    m: MagicMock = mock_open(read_data=b"")
    with pytest.raises(ValueError):
        with patch("builtins.open", m):
            h = Forwarh.read(ARQ_TESTE)


def test_eq_forwarh():
    h1 = Forwarh.read(ARQ_TESTE)
    h2 = Forwarh.read(ARQ_TESTE)
    assert h1 == h2


def test_atributos_forwarh():
    h1 = Forwarh.read(ARQ_TESTE)
    assert (
        h1.nome_caso
        == "Backtest Preliminar CPAMP 2022-2023 - Hibrido NEWAVE 05/2021"
    )

    assert h1.numero_rees == 12
    assert h1.numero_submercados == 4
    assert h1.numero_total_submercados == 5
    assert h1.numero_series_gravadas == 2000
    assert h1.numero_aberturas == 20
    assert h1.numero_estagios_estudo == 60
    assert h1.intervalo_series_gravadas == 1
    assert h1.numero_classes_termicas_submercados == [34, 19, 43, 24]
    assert h1.numero_patamares_deficit == 1
    assert h1.tamanho_registro_arquivo_forward == 41264
    assert h1.numero_registros_arquivo_forward == 2000
    assert h1.numero_registros_necessarios_estagio == 1
    assert h1.ano_inicio_estudo == 2021
    assert h1.ano_inicio_historico_vazoes == 1
    assert h1.numero_anos_descontar_historico_vazoes == 0
    assert h1.numero_estagios_ano == 12
    assert h1.mes_inicio_estudo == 5
    assert h1.mes_inicio_pre_estudo == 1
    assert h1.numero_estagios_pre_estudo == 0
    assert h1.ordem_maxima_parp == 12
    assert h1.ano_inicio_series_historicas_simuladas == [0] * 100
    assert h1.numero_anos_historico_vazoes == 89
    assert h1.numero_patamares_carga == 3
    assert h1.simulacao_final_individualizada == 1



from inewave.newave.modelos.vazpast import BlocoVazPast

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Vazpast(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às
    vazões anteriores ao período de planejamento.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que são usadas juntos das contidas no arquivo `vazoes.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoVazPast]

    @property
    def tendencia(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a tendência hidrológica por UHE.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - mes (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoVazPast)
        if isinstance(b, BlocoVazPast):
            return b.data
        return None

    @tendencia.setter
    def tendencia(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoVazPast)
        if isinstance(b, BlocoVazPast):
            b.data = df



from inewave.nwlistop.modelos.cdef import CdefAnos

from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
    Submercado,
)


class Cdef(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes ao custo de déficit
    de cada estágio em cada série.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cdef00x.out`.
    """

    BLOCKS = [CdefAnos, Submercado]



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.ghtotm import GHAnos


class Ghtotm(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à geração hidraulica total
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghtotm00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        GHAnos,
    ]



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.corteolm import CorteolmAnos


class Corteolm(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes ao corte de geração eólica.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `corteolm00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Submercado,
        CorteolmAnos,
    ]



from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.hidr import RegistroUHEHidr
from inewave.config import MESES_ABREV
import pandas as pd  # type: ignore


from typing import TypeVar, List, Optional, Union, IO


class Hidr(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro das
    usinas hidroelétricas.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroUHEHidr]
    STORAGE = "BINARY"

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__df: Optional[pd.DataFrame] = None

    def write(self, to: Union[str, IO], *args, **kwargs):
        self.__atualiza_registros()
        super().write(to, *args, **kwargs)

    def __monta_df_de_registros(self) -> Optional[pd.DataFrame]:
        registros: List[RegistroUHEHidr] = [
            r for r in self.data.of_type(RegistroUHEHidr)
        ]
        if len(registros) == 0:
            return None
        df = pd.DataFrame(
            columns=[
                "nome_usina",
                "posto",
                "submercado",
                "empresa",
                "codigo_usina_jusante",
                "desvio",
                "volume_minimo",
                "volume_maximo",
                "volume_vertedouro",
                "volume_desvio",
                "cota_minima",
                "cota_maxima",
                *[f"a{i}_volume_cota" for i in range(5)],
                *[f"a{i}_cota_area" for i in range(5)],
                *[f"evaporacao_{m}" for m in MESES_ABREV],
                "numero_conjuntos_maquinas",
                *[f"maquinas_conjunto_{i}" for i in range(1, 6)],
                *[f"potencia_nominal_conjunto_{i}" for i in range(1, 6)],
                *[f"queda_nominal_conjunto_{i}" for i in range(1, 6)],
                *[f"vazao_nominal_conjunto_{i}" for i in range(1, 6)],
                "produtibilidade_especifica",
                "perdas",
                "numero_polinomios_jusante",
                *[f"a{i}_jusante_1" for i in range(5)],
                *[f"a{i}_jusante_2" for i in range(5)],
                *[f"a{i}_jusante_3" for i in range(5)],
                *[f"a{i}_jusante_4" for i in range(5)],
                *[f"a{i}_jusante_5" for i in range(5)],
                *[f"a{i}_jusante_6" for i in range(5)],
                *[f"referencia_jusante_{i}" for i in range(1, 7)],
                "canal_fuga_medio",
                "influencia_vertimento_canal_fuga",
                "fator_carga_maximo",
                "fator_carga_minimo",
                "vazao_minima_historica",
                "numero_unidades_base",
                "tipo_turbina",
                "representacao_conjunto",
                "teif",
                "ip",
                "tipo_perda",
                "data",
                "observacao",
                "volume_referencia",
                "tipo_regulacao",
            ]
        )
        for i, r in enumerate(registros):
            df.loc[i + 1] = [
                r.nome,
                r.posto,
                r.subsistema,
                r.empresa,
                r.jusante,
                r.desvio,
                r.volume_minimo,
                r.volume_maximo,
                r.volume_vertedouro,
                r.volume_desvio,
                r.cota_minima,
                r.cota_maxima,
                *(r.polinomio_volume_cota),
                *(r.polinomio_cota_area),
                *(r.evaporacao),
                r.numero_conjuntos_maquinas,
                *(r.numero_maquinas_conjunto),
                *(r.potef_conjunto),
                *(r.hef_conjunto),
                *(r.qef_conjunto),
                r.produtibilidade_especifica,
                r.perdas,
                r.numero_polinomios_jusante,
                *(r.polinomios_jusante),
                r.canal_fuga_medio,
                r.influencia_vertimento_canal_fuga,
                r.fator_carga_maximo,
                r.fator_carga_minimo,
                r.vazao_minima_historica,
                r.numero_unidades_base,
                r.tipo_turbina,
                r.representacao_conjunto,
                r.teif,
                r.ip,
                r.tipo_perda,
                r.data_referencia,
                r.observacao,
                r.volume_referencia,
                r.tipo_regulacao,
            ]
        df.index.name = "codigo_usina"

        df = df.astype(
            {
                "nome_usina": str,
                "posto": int,
                "submercado": int,
                "empresa": int,
                "codigo_usina_jusante": int,
                "desvio": int,
                "volume_minimo": float,
                "volume_maximo": float,
                "volume_vertedouro": float,
                "volume_desvio": float,
                "cota_minima": float,
                "cota_maxima": float,
                "produtibilidade_especifica": float,
                "perdas": float,
                "canal_fuga_medio": float,
                "influencia_vertimento_canal_fuga": int,
                "fator_carga_maximo": float,
                "fator_carga_minimo": float,
                "vazao_minima_historica": int,
                "numero_unidades_base": int,
                "tipo_turbina": int,
                "representacao_conjunto": int,
                "teif": float,
                "ip": float,
                "tipo_perda": int,
                "data": str,
                "observacao": str,
                "volume_referencia": float,
                "tipo_regulacao": str,
            },
        )
        return df

    def __atualiza_registros(self):
        registros: List[RegistroUHEHidr] = [r for r in self.data][1:]
        for (_, linha), r in zip(self.cadastro.iterrows(), registros):
            r.nome = linha["nome_usina"]
            r.posto = linha["posto"]
            r.subsistema = linha["submercado"]
            r.empresa = linha["empresa"]
            r.jusante = linha["codigo_usina_jusante"]
            r.desvio = linha["desvio"]
            r.volume_minimo = linha["volume_minimo"]
            r.volume_maximo = linha["volume_maximo"]
            r.volume_vertedouro = linha["volume_vertedouro"]
            r.volume_desvio = linha["volume_desvio"]
            r.cota_minima = linha["cota_minima"]
            r.cota_maxima = linha["cota_maxima"]
            r.polinomio_volume_cota = linha[
                [f"a{i}_volume_cota" for i in range(5)]
            ].tolist()
            r.polinomio_cota_area = linha[
                [f"a{i}_cota_area" for i in range(5)]
            ].tolist()
            r.evaporacao = linha[
                [f"evaporacao_{m}" for m in MESES_ABREV]
            ].tolist()
            r.numero_conjuntos_maquinas = linha["numero_conjuntos_maquinas"]
            r.potef_conjunto = linha[
                [f"potencia_nominal_conjunto_{i}" for i in range(1, 6)]
            ].tolist()
            r.hef_conjunto = linha[
                [f"queda_nominal_conjunto_{i}" for i in range(1, 6)]
            ].tolist()
            r.qef_conjunto = linha[
                [f"vazao_nominal_conjunto_{i}" for i in range(1, 6)]
            ].tolist()
            r.produtibilidade_especifica = linha["produtibilidade_especifica"]
            r.perdas = linha["perdas"]
            r.numero_polinomios_jusante = linha["numero_polinomios_jusante"]
            r.polinomios_jusante = linha[
                [f"a{i}_jusante_1" for i in range(5)]
                + [f"a{i}_jusante_2" for i in range(5)]
                + [f"a{i}_jusante_3" for i in range(5)]
                + [f"a{i}_jusante_4" for i in range(5)]
                + [f"a{i}_jusante_5" for i in range(5)]
                + [f"a{i}_jusante_6" for i in range(5)]
                + [f"referencia_jusante_{i}" for i in range(1, 7)]
            ].tolist()
            r.canal_fuga_medio = linha["canal_fuga_medio"]
            r.influencia_vertimento_canal_fuga = linha[
                "influencia_vertimento_canal_fuga"
            ]
            r.fator_carga_maximo = linha["fator_carga_maximo"]
            r.fator_carga_minimo = linha["fator_carga_minimo"]
            r.vazao_minima_historica = linha["vazao_minima_historica"]
            r.numero_unidades_base = linha["numero_unidades_base"]
            r.tipo_turbina = linha["tipo_turbina"]
            r.representacao_conjunto = linha["representacao_conjunto"]
            r.teif = linha["teif"]
            r.ip = linha["ip"]
            r.tipo_perda = linha["tipo_perda"]
            r.data_referencia = linha["data"]
            r.observacao = linha["observacao"]
            r.volume_referencia = linha["volume_referencia"]
            r.tipo_regulacao = linha["tipo_regulacao"]

    @property
    def cadastro(self) -> pd.DataFrame:
        """
        Obtém a tabela com os dados cadastrais existentes no arquivo
        binário.

        - nome_usina (`str`): nome da usina (12 caracteres)
        - posto (`int`): posto de vazão natural da usina
        - submercado (`int`): submercado da usina
        - empresa (`int`): agente responsável pela usina
        - codigo_usina_jusante (`int`): posto à jusante da usina
        - desvio (`float`): TODO
        - volume_minimo (`float`): volume mínimo da usina (hm3)
        - volume_maximo (`float`): volume máximo da usina (hm3)
        - volume_vertedouro (`float`): volume do vertedouro da usina (hm3)
        - volume_desvio (`float`): TODO
        - cota_minima (`float`): cota mínima da usina (m)
        - cota_maxima (`float`): cota máxima da usina (m)
        - a[0-4]_volume_cota (`float`): coeficientes do polinômio volume-cota
        - a[0-4]_cota_area (`float`): coeficientes do polinômio cota-área
        - evaporacao_[JAN..DEZ] (`float`): coeficientes de evaporação (mm)
        - numero_conjuntos_maquinas (`int`): número de conjuntos de máquinas
        - maquinas_conjunto_[1-5] (`int`): máquinas por conjunto
        - potencia_nominal_conjunto_[1-5] (`float`): potência das máquinas (MWmed)
        - queda_nominal_conjunto_[1-5]: alturas nominais de queda por conjunto (m)
        - vazao_nominal_conjunto_[1-5]: vazões nominais por conjunto (m3/s)
        - produtibilidade_especifica (`float`): produtibilidade específica
        - perdas (`float`): perdas da usina
        - numero_polinomios_jusante (`int`): número de polinômios de jusante
        - a[0-4]_jusante_[1-6] (`float`): coeficientes de cada polinjus
        - referencia_jusante_[1-6] (`float`): coeficientes do polinjus de referência
        - canal_fuga_medio (`float`): cota média do canal de fuga (m)
        - influencia_vertimento_canal_fuga (`int`): TODO (0 ou 1)
        - fator_carga_maximo (`float`): TODO (%)
        - fator_carga_minimo (`float`): TODO (%)
        - vazao_minima_historica (`float`): vazão mínima da usina (m3/s)
        - numero_unidades_base (`int`): TODO
        - tipo_turbina (`int`): TODO
        - representacao_conjunto (`int`): TODO
        - teif (`float`): TODO (%)
        - ip (`float`): TODO (%)
        - tipo_perda (`int`): TODO
        - data (`str`): DD-MM-AA
        - observacao (`str`): observação qualquer sobre a usina
        - volume_referencia (`float`): TODO (hm3)
        - tipo_regulacao (`str`): D, S ou M

        :return: A tabela com os dados cadastrais
        :rtype: pd.DataFrame | None
        """
        if self.__df is None:
            self.__df = self.__monta_df_de_registros()
        return self.__df

    @cadastro.setter
    def cadastro(self, df: pd.DataFrame):
        self.__df = df



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField


class RegistroPostoVentoCadastro(Register):
    """
    Registro que contém um cadastro de um posto de vento.
    """

    IDENTIFIER = "POSTO-VENTO-CAD"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_posto(self) -> Optional[int]:
        return self.data[0]

    @codigo_posto.setter
    def codigo_posto(self, c: int):
        self.data[0] = c

    @property
    def nome_posto(self) -> Optional[str]:
        return self.data[1]

    @nome_posto.setter
    def nome_posto(self, n: str):
        self.data[1] = n


class RegistroPEEPostoVento(Register):
    """
    Registro que contém um mapeamento de PEE para posto de vento
    """

    IDENTIFIER = "PEE-POSTO"
    IDENTIFIER_DIGITS = 9
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def codigo_posto(self) -> Optional[int]:
        return self.data[1]

    @codigo_posto.setter
    def codigo_posto(self, c: int):
        self.data[1] = c



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.restricaoenergia import (
    RegistroRHE,
    RegistroRHEHorizPer,
    RegistroRHELsLPPEarmi,
)


class RestricaoEnergia(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro
    das restrições lineares por partes no domínio de energia (REE).
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroRHELsLPPEarmi,
        RegistroRHEHorizPer,
        RegistroRHE,
    ]

    def rhe(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
    ) -> Optional[Union[RegistroRHE, List[RegistroRHE]]]:
        """
        Obtém um registro que cadastra uma usina restrição linear por
        partes de energia (REE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHE` |
            list[:class:`RegistroRHE`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHE,
            codigo_restricao=codigo_restricao,
            formula=formula,
        )

    def rhe_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[Union[RegistroRHEHorizPer, List[RegistroRHEHorizPer]]]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição linear por partes de energia.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicial: data inicial de validade da restrição
        :type data_inicial: datetime | None
        :param data_final: data final de validade da restrição
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHEHorizPer` |
            list[:class:`RegistroRHEHorizPer`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHEHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def rhe_ls_lpp_earmi(
        self,
        codigo_restricao: Optional[int] = None,
        indice_reta: Optional[int] = None,
        coeficiente_angular: Optional[float] = None,
        coeficiente_linear: Optional[float] = None,
    ) -> Optional[Union[RegistroRHELsLPPEarmi, List[RegistroRHELsLPPEarmi]]]:
        """
        Obtém um registro que cadastra as retas da restrição linear
        por partes de energia.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param indice_reta: índice da reta definida
        :type indice_reta: int | None
        :param coeficiente_angular: coeficiente angular da reta
        :type coeficiente_angular: float | None
        :param coeficiente_linear: coeficiente linear da reta
        :type coeficiente_linear: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHELsLPPEarmi` |
            list[:class:`RegistroRHELsLPPEarmi`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHELsLPPEarmi,
            codigo_restricao=codigo_restricao,
            indice_reta=indice_reta,
            coeficiente_angular=coeficiente_angular,
            coeficiente_linear=coeficiente_linear,
        )



from inewave.newave.modelos.vazaof import SecaoDadosVazaof
from inewave.newave.vazaof import Vazaof


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TESTE = "./tests/mocks/arquivos/vazaof.dat"

NUM_FORWARDS = 2
NUM_UHES = 1
NUM_ESTAGIOS = 16
NUM_ESTAGIOS_TH = 12
NUM_ENTRADAS = NUM_FORWARDS * NUM_UHES * (NUM_ESTAGIOS_TH + NUM_ESTAGIOS)


def test_secao_vazao():
    r = SecaoDadosVazaof()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(
            fp,
            numero_forwards=NUM_FORWARDS,
            numero_uhes=NUM_UHES,
            numero_estagios=NUM_ESTAGIOS,
            numero_estagios_th=NUM_ESTAGIOS_TH,
        )

    assert len(r.data) == NUM_ENTRADAS


def test_atributos_encontrados_vazao():
    h = Vazaof.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_uhes=NUM_UHES,
        numero_estagios=NUM_ESTAGIOS,
        numero_estagios_th=NUM_ESTAGIOS_TH,
    )
    assert h.series is not None
    assert h.series.isna().sum().sum() == 0


def test_atributos_nao_encontrados_vazao():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h = Vazaof.read(
            ARQ_TESTE,
            numero_forwards=NUM_FORWARDS,
            numero_uhes=NUM_UHES,
            numero_estagios=NUM_ESTAGIOS,
            numero_estagios_th=NUM_ESTAGIOS_TH,
        )
        assert h.series.isna().sum().sum() == NUM_ENTRADAS


def test_eq_vazao():
    h1 = Vazaof.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_uhes=NUM_UHES,
        numero_estagios=NUM_ESTAGIOS,
        numero_estagios_th=NUM_ESTAGIOS_TH,
    )
    h2 = Vazaof.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_uhes=NUM_UHES,
        numero_estagios=NUM_ESTAGIOS,
        numero_estagios_th=NUM_ESTAGIOS_TH,
    )
    assert h1 == h2



from inewave.newave.modelos.cortesh import SecaoDadosCortesh
from inewave.newave.cortesh import Cortesh

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch
import pytest

ARQ_TESTE = "./tests/mocks/arquivos/cortesh.dat"


def test_secao_dados_cortesh():
    r = SecaoDadosCortesh()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(fp, storage="BINARY")

    assert len(r.data) == 23865
    assert r.versao_newave == 281200
    assert r.tamanho_corte == 17568
    assert r.tamanho_estado == 0
    assert r.numero_rees == 12
    assert r.numero_estagios_pre == 0
    assert r.numero_estagios_estudo == 60
    assert r.numero_estagios_pos == 60
    assert r.numero_estagios_ano == 12
    assert r.numero_configuracoes == 56
    assert r.numero_forwards == 200
    assert r.numero_patamares == 3
    assert r.ano_inicio_estudo == 2022
    assert r.mes_inicio_estudo == 5
    assert r.lag_maximo_gnl == 2
    assert r.mecanismo_aversao == 4
    assert r.numero_submercados == 4
    assert r.numero_total_submercados == 5
    assert r.usa_curva_aversao == 1
    assert r.usa_sar == 0
    assert r.usa_cvar == 1
    assert r.considera_no_zero_calculo_zinf == 0
    assert r.mes_agregacao == 17
    assert r.numero_maximo_uhes == 164
    assert r.considera_afluencia_anual == 0
    assert r.versao_nao_oficial == "CPAMP"
    assert r.tipo_penalizacao_curva == 1
    assert r.mes_penalizacao_curva == 11
    assert r.opcao_parpa == 3
    assert r.tipo_agregacao_caso == 0
    assert r.estagio_individualizado_inicial == 5
    assert r.estagio_individualizado_final == 16
    assert r.estagio_agregado_inicial == 17
    assert r.estagio_agregado_final == 120
    assert r.tamanho_registro_individualizado == 17568
    assert r.tamanho_registro_agregado == 1664
    assert r.ultimo_registro_cortes_estagio.shape == (120, 3)
    assert len(r.ordens_modelos_parp) == 1440
    assert len(r.configuracoes) == 120
    assert len(r.duracoes_patamares) == 180
    assert r.iteracao_atual == 30
    assert len(r.penalidade_violacao_curva) == 12
    assert len(r.curva_aversao) == 720
    assert r.flag_cvar == 1
    assert len(r.alfa_cvar) == 84
    assert len(r.lambda_cvar) == 84
    assert r.numero_uhes == 164
    assert len(r.codigos_uhes) == 164
    assert len(r.codigo_interno_ree_uhes) == 164
    assert len(r.rees_por_submercado) == 4
    assert len(r.codigos_internos_rees_por_submercado) == 12
    assert len(r.nomes_rees_submercados) == 16
    assert len(r.codigos_rees_submercados) == 16
    assert r.numero_coeficientes_derivacao_inexata_parpa == 7
    assert r.constante_numero_rees == 12
    assert r.constante_numero_rees_ordem_maxima_parp == 144
    assert r.constante_numero_submercados_patamares_carga_lag_maximo_gnl == 24
    assert r.constante_numero_maximo_uhes == 164
    assert r.constante_numero_maximo_uhes_ordem_maxima_parp == 1968
    assert len(r.numero_uhes_estagios_individualizados) == 17
    assert r.dados_uhes.shape == (164, 10)
    assert r.dados_submercados.shape == (5, 3)


def test_atributos_encontrados_cortesh():
    h = Cortesh.read(ARQ_TESTE)
    assert h.versao_newave is not None


def test_atributos_nao_encontrados_cortesh():
    m: MagicMock = mock_open(read_data=b"")
    with pytest.raises(ValueError):
        with patch("builtins.open", m):
            h = Cortesh.read(ARQ_TESTE)


def test_eq_cortesh():
    h1 = Cortesh.read(ARQ_TESTE)
    h2 = Cortesh.read(ARQ_TESTE)
    assert h1 == h2


def test_atributos_cortesh():
    h1 = Cortesh.read(ARQ_TESTE)
    assert h1.versao_newave == 281200
    assert h1.tamanho_corte == 17568
    assert h1.tamanho_estado == 0
    assert h1.numero_rees == 12
    assert h1.numero_estagios_pre == 0
    assert h1.numero_estagios_estudo == 60
    assert h1.numero_estagios_pos == 60
    assert h1.numero_estagios_ano == 12
    assert h1.numero_configuracoes == 56
    assert h1.numero_forwards == 200
    assert h1.numero_patamares == 3
    assert h1.ano_inicio_estudo == 2022
    assert h1.mes_inicio_estudo == 5
    assert h1.lag_maximo_gnl == 2
    assert h1.numero_submercados == 4
    assert h1.numero_total_submercados == 5
    assert h1.mes_agregacao == 17
    assert h1.numero_maximo_uhes == 164
    assert h1.considera_afluencia_anual == 0
    assert h1.tipo_agregacao_caso == 0
    assert h1.estagio_individualizado_inicial == 5
    assert h1.estagio_individualizado_final == 16
    assert h1.estagio_agregado_inicial == 17
    assert h1.estagio_agregado_final == 120
    assert h1.tamanho_registro_individualizado == 17568
    assert h1.tamanho_registro_agregado == 1664
    assert h1.ultimo_registro_cortes_estagio.shape == (120, 3)
    assert h1.dados_uhes.shape == (164, 10)
    assert h1.dados_submercados.shape == (5, 3)



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from typing import List, IO


class BlocoVarreduraShist(Section):
    """
    Bloco de informações de varredura existente no arquivo `shist.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__line = Line(
            [
                IntegerField(3, 0),
                IntegerField(4, 4),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoVarreduraShist):
            return False
        bloco: BlocoVarreduraShist = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        # Salta as linhas de cabeçalhos
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data = self.__line.read(file.readline())

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, list):
            raise ValueError("Dados do shist.dat não foram lidos com sucesso")

        file.write(self.__line.write(self.data))


class BlocoSeriesSimulacaoShist(Section):
    """
    Bloco de informações das séries para a simulação
    existente no arquivo `shist.dat`.
    """

    END_PATTERN = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__line = Line(
            [
                IntegerField(4, 0),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSeriesSimulacaoShist):
            return False
        bloco: BlocoSeriesSimulacaoShist = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        # Salta as linhas de cabeçalhos
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data: List[int] = []
        while True:
            linha = file.readline()
            if (
                BlocoSeriesSimulacaoShist.END_PATTERN in linha[:4]
                or len(linha) < 2
            ):
                break

            self.data.append(self.__line.read(linha)[0])

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, list):
            raise ValueError("Dados do shist.dat não foram lidos com sucesso")

        for s in self.data:
            file.write(self.__line.write([s]))

        file.write(BlocoSeriesSimulacaoShist.END_PATTERN + "\n")



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicaconfiguracao import (
    RegistroEolicaConfiguracao,
    RegistroPEEConfiguracaoPeriodo,
)


class EolicaConfiguracao(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    das usinas eólicas.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroEolicaConfiguracao, RegistroPEEConfiguracaoPeriodo]

    def eolica_configuracao(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        estado_operacao: Optional[str] = None,
    ) -> Optional[
        Union[
            RegistroEolicaConfiguracao,
            List[RegistroEolicaConfiguracao],
        ]
    ]:
        """
        Obtém um registro que contém a configuração de operação para
        uma usina eólica em um período.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param data_inicial: período de início da configuração
        :type data_inicial: datetime | None
        :param data_final: período de fim da configuração
        :type data_final: datetime | None
        :param estado_operacao: tipo de operação
        :type estado_operacao: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaConfiguracao` |
            list[:class:`RegistroEolicaConfiguracao`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaConfiguracao,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            estado_operacao=estado_operacao,
        )

    def pee_config_per(
        self,
        codigo_pee: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        estado_operacao: Optional[str] = None,
    ) -> Optional[
        Union[
            RegistroPEEConfiguracaoPeriodo,
            List[RegistroPEEConfiguracaoPeriodo],
        ]
    ]:
        """
        Obtém um registro que contém a configuração de operação para
        um PEE em um período.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param data_inicial: período de início da configuração
        :type data_inicial: datetime | None
        :param data_final: período de fim da configuração
        :type data_final: datetime | None
        :param estado_operacao: tipo de operação
        :type estado_operacao: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEConfiguracaoPeriodo` |
            list[:class:`RegistroPEEConfiguracaoPeriodo`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPEEConfiguracaoPeriodo,
            codigo_pee=codigo_pee,
            data_inicial=data_inicial,
            data_final=data_final,
            estado_operacao=estado_operacao,
        )



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.vghmin import VghminAnos


class Vghminsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação da meta de geração
    hidraulica mínima por patamar, para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `vghminsin.out`.

    """

    BLOCKS = [
        VghminAnos,
    ]



from cfinterface.components.block import Block
from cfinterface.components.line import Line

from typing import IO, List, Dict
import pandas as pd  # type: ignore


class TabelaCSV(Block):
    """
    Bloco para ler uma tabela com separadores CSV fornecidos
    a partir de um modelo de linha, para arquivos de saída do NEWAVE.
    """

    BEGIN_PATTERN = "-----;------;"
    LINE_MODEL = Line([])
    COLUMN_NAMES: List[str] = []
    END_PATTERN = ""

    def _monta_df(self, dados: dict) -> pd.DataFrame:
        return pd.DataFrame(data=dados, columns=self.__class__.COLUMN_NAMES)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TabelaCSV):
            return False
        else:
            if not all(
                [type(self.data) is pd.DataFrame, type(o.data) is pd.DataFrame]
            ):
                return False
            else:
                return self.data.equals(o.data)

    def read(self, file: IO, *args, **kwargs):
        if len(self.__class__.LINE_MODEL.fields) != len(
            self.__class__.COLUMN_NAMES
        ):
            n_linha = len(self.__class__.LINE_MODEL.fields)
            n_cols = len(self.__class__.COLUMN_NAMES)
            raise RuntimeError(
                f"Número de colunas ({n_cols}) diferente do"
                + f" número de campos da linha ({n_linha})"
            )
        # Espera o fim do cabeçalho
        linha = file.readline()
        while True:
            linha = file.readline()
            if self.__class__.BEGIN_PATTERN in linha:
                break
            elif len(linha) < 3:
                return
        # Lê a tabela
        dados: Dict[str, List] = {c: [] for c in self.__class__.COLUMN_NAMES}
        while True:
            linha = file.readline()
            if (self.__class__.BEGIN_PATTERN in linha) or (len(linha) < 3):
                self.data = self._monta_df(dados)
                return
            dados_linha = self.__class__.LINE_MODEL.read(linha)
            for i, c in enumerate(self.__class__.COLUMN_NAMES):
                dados[c].append(dados_linha[i])



from inewave.newave.modelos.blocos.versaomodelo import VersaoModelo
from inewave.newave.modelos.avl_desvfpha_s import TabelaAvlDesvFphaS

from cfinterface.files.blockfile import BlockFile
from typing import Optional, TypeVar
import pandas as pd  # type: ignore


class AvlDesvFphaS(BlockFile):
    """
    Arquivo com os desvios da função de produção no plano de
    vazão vertida (S).
    """

    BLOCKS = [VersaoModelo, TabelaAvlDesvFphaS]
    ENCODING = "iso-8859-1"

    T = TypeVar("T")

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__df_completo: Optional[pd.DataFrame] = None

    @property
    def tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - volume_armazenado_percentual (`float`)
        - vazao_turbinada_m3s (`float`)
        - vazao_vertida_m3s (`float`)
        - desvio_percentual (`float`)

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        if self.__df_completo is None:
            tabelas = self.data.of_type(TabelaAvlDesvFphaS)
            self.__df_completo = pd.DataFrame()
            for t in tabelas:
                self.__df_completo = pd.concat(
                    [self.__df_completo, t.data], ignore_index=True
                )
        return self.__df_completo

    @property
    def versao(self) -> Optional[str]:
        """
        A versão do modelo utilizada para executar o caso.

        :return: A versão do modelo
        :rtype: str | None
        """
        b = self.data.get_blocks_of_type(VersaoModelo)
        if isinstance(b, VersaoModelo):
            return b.data
        return None



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicageracao import (
    RegistroEolicaGeracaoPeriodo,
    RegistroEolicaGeracaoPatamar,
    RegistroPEEGeracaoPatamar,
)


class EolicaGeracao(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao histórico
    de geração das usinas eólicas.
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroEolicaGeracaoPeriodo,
        RegistroEolicaGeracaoPatamar,
        RegistroPEEGeracaoPatamar,
    ]

    def eolica_geracao_periodo(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        geracao: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaGeracaoPeriodo,
            List[RegistroEolicaGeracaoPeriodo],
        ]
    ]:
        """
        Obtém um registro que contém o valor de geração para uma usina
        eólica durante um período.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicial: data de início do histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do histórico
        :type data_final: datetime | None
        :param geracao: valor de geração
        :type geracao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaGeracaoPeriodo` |
            list[:class:`RegistroEolicaGeracaoPeriodo`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaGeracaoPeriodo,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            geracao=geracao,
        )

    def eolica_geracao_profundidade_periodo_patamar(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        indice_patamar: Optional[int] = None,
        profundidade: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaGeracaoPatamar,
            List[RegistroEolicaGeracaoPatamar],
        ]
    ]:
        """
        Obtém um registro que contém a profundidade de um patamar
        de geração para um período.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicial: data de início do registro histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do registro histórico
        :type data_final: datetime | None
        :param indice_patamar: patamar de geração eólica
        :type indice_patamar: int | None
        :param profundidade: profundidade do patamar de geração
        :type profundidade: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaGeracaoPatamar` |
            list[:class:`RegistroEolicaGeracaoPatamar`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaGeracaoPatamar,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            indice_patamar=indice_patamar,
            profundidade=profundidade,
        )

    def pee_ger_prof_per_pat(
        self,
        codigo_pee: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        indice_patamar: Optional[int] = None,
        profundidade: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroPEEGeracaoPatamar,
            List[RegistroPEEGeracaoPatamar],
        ]
    ]:
        """
        Obtém um registro que contém a profundidade de um patamar
        de geração para um período, para um PEE.

        :param codigo_pee: código do PEE
        :type codigo_pee: int | None
        :param data_inicial: data de início do registro histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do registro histórico
        :type data_final: datetime | None
        :param indice_patamar: patamar de geração eólica
        :type indice_patamar: int | None
        :param profundidade: profundidade do patamar de geração
        :type profundidade: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEGeracaoPatamar` |
            list[:class:`RegistroPEEGeracaoPatamar`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPEEGeracaoPatamar,
            codigo_pee=codigo_pee,
            data_inicial=data_inicial,
            data_final=data_final,
            indice_patamar=indice_patamar,
            profundidade=profundidade,
        )



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.vturuh import VturAnos


class Vturuh(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes às vazões turbinadas por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vturuh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        VturAnos,
    ]



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicahistorico import (
    RegistroEolicaHistoricoVentoHorizonte,
    RegistroEolicaHistoricoVento,
    RegistroHistoricoVentoHorizonte,
    RegistroHistoricoVento,
)


class EolicaHistorico(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao histórico
    de ventos das usinas eólicas.
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroEolicaHistoricoVentoHorizonte,
        RegistroEolicaHistoricoVento,
        RegistroHistoricoVentoHorizonte,
        RegistroHistoricoVento,
    ]

    def eolica_historico_vento_horizonte(
        self,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[
        Union[
            RegistroEolicaHistoricoVentoHorizonte,
            List[RegistroEolicaHistoricoVentoHorizonte],
        ]
    ]:
        """
        Obtém um registro que contém o horizonte de um histórico.

        :param data_inicial: data de início do histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do histórico
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaHistoricoVentoHorizonte` |
            list[:class:`RegistroEolicaHistoricoVentoHorizonte`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaHistoricoVentoHorizonte,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def eolica_historico_vento(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        velocidade: Optional[float] = None,
        direcao: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaHistoricoVento,
            List[RegistroEolicaHistoricoVento],
        ]
    ]:
        """
        Obtém um registro que contém a o valor de vento para um
        intervalo do histórico.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicial: data de início do registro histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do registro histórico
        :type data_final: datetime | None
        :param velocidade: velocidade do vento
        :type velocidade: float | None
        :param direcao: direção do vento
        :type direcao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaHistoricoVento` |
            list[:class:`RegistroEolicaHistoricoVento`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaHistoricoVento,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            velocidade=velocidade,
            direcao=direcao,
        )

    def vento_hist_horiz(
        self,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[
        Union[
            RegistroHistoricoVentoHorizonte,
            List[RegistroHistoricoVentoHorizonte],
        ]
    ]:
        """
        Obtém um registro que contém o horizonte de um histórico.

        :param data_inicial: data de início do histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do histórico
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroHistoricoVentoHorizonte` |
            list[:class:`RegistroHistoricoVentoHorizonte`] | None
        """
        return self.data.get_registers_of_type(
            RegistroHistoricoVentoHorizonte,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def vento_hist(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        velocidade: Optional[float] = None,
        direcao: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroHistoricoVento,
            List[RegistroHistoricoVento],
        ]
    ]:
        """
        Obtém um registro que contém a o valor de vento para um
        intervalo do histórico.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicial: data de início do registro histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do registro histórico
        :type data_final: datetime | None
        :param velocidade: velocidade do vento
        :type velocidade: float | None
        :param direcao: direção do vento
        :type direcao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroHistoricoVento` |
            list[:class:`RegistroHistoricoVento`] | None
        """
        return self.data.get_registers_of_type(
            RegistroHistoricoVento,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            velocidade=velocidade,
            direcao=direcao,
        )



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional

import pandas as pd  # type: ignore

from inewave.newave.modelos.ree import (
    BlocoReesSubmercados,
    BlocoFicticiasIndividualizado,
)


class Ree(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    dos REEs.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [
        BlocoReesSubmercados,
        BlocoFicticiasIndividualizado,
    ]

    @property
    def rees(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os REES e os submercados

        - codigo (`int`)
        - nome (`str`)
        - submercado (`int`)
        - mes_fim_individualizado (`int`)
        - ano_fim_individualizado (`int`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoReesSubmercados)
        if isinstance(b, BlocoReesSubmercados):
            return b.data
        return None

    @rees.setter
    def rees(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoReesSubmercados)
        if isinstance(b, BlocoReesSubmercados):
            b.data = df

    @property
    def remocao_ficticias(self) -> Optional[int]:
        """
        Opção de remover usinas fictícias nos períodos individualizados.

        :return: O valor do campo.
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoFicticiasIndividualizado)
        if isinstance(b, BlocoFicticiasIndividualizado):
            return b.data[1]
        return None

    @remocao_ficticias.setter
    def remocao_ficticias(self, d: int):
        b = self.data.get_sections_of_type(BlocoFicticiasIndividualizado)
        if isinstance(b, BlocoFicticiasIndividualizado):
            b.data[1] = d
